import os
import pickle
import re
from collections import Counter
from functools import lru_cache
import warnings

import numpy as np
from nltk.tokenize import word_tokenize
import torch
from torchtext.vocab import Vectors, GloVe

from . import constants

word_detector = re.compile('\w')


class VocabModel(object):
    """Vocab model builder.

    Parameters
    ----------
    data_set: iterable
        A list of instances where each instance is a list of str.
    tokenizer: function, optional
        Word tokenization function, default: ``nltk.tokenize.word_tokenize``.
    max_word_vocab_size: int, optional
        Maximal word vocab size, default: ``None``.
    min_word_vocab_freq: int, optional
        Minimal word vocab frequency, default: ``1``.
    pretrained_word_emb_name: str, optional, default="840B"
        The name of pretrained word embedding in ``torchtext``.
        If it is set ``None``, we will randomly set the initial word embedding values.
    pretrained_word_emb_url: str optional, default: ``None``
        The url for downloading pretrained word embedding.
        Note that we only prepare the default ``url`` for English with ``pretrained_word_emb_name`` as ``"42B"``, ``"840B"``, 'twitter.27B' and '6B'.
    target_pretrained_word_emb_name: str, optional, default=None
        The name of pretrained word embedding in ``torchtext`` for target language.
        If it is set ``None``, we will use ``pretrained_word_emb_name``.
    target_pretrained_word_emb_url: str optional, default: ``None``
        The url for downloading pretrained word embedding for target language.
        Note that we only prepare the default ``url`` for English with ``pretrained_word_emb_name`` as ``"42B"``, ``"840B"``, 'twitter.27B' and '6B'.
    pretrained_word_emb_cache_dir: str, optional, default: ``".vector_cache/"``
        The path of directory saving the temporary word embedding file.
    word_emb_size: int, optional
        Word embedding size, default: ``None``.
    share_vocab : boolean
        Specify whether to share vocab between input and output text, default: ``True``.

    Examples
    -------
    # Build a vocab model from scratch
    >>> vocab_model = VocabModel([['I like nlp.', 'Same here!'],
                        ['I like graph.', 'Same here!']],
                        max_word_vocab_size=None,
                        min_word_vocab_freq=1,
                        word_emb_size=300)
    >>> print(vocab_model.word_vocab.get_vocab_size())

    # Restore a vocab model from disk if exists or build one from scratch.
    >>> vocab_model = VocabModel.build('vocab_model.pkl', [['I like nlp.', 'Same here!'],
                            ['I like graph.', 'Same here!']],
                            max_word_vocab_size=None,
                            min_word_vocab_freq=1,
                            word_emb_size=300)
    >>> print(vocab_model.word_vocab.get_vocab_size())
    """

    def __init__(self, data_set=None,
                 tokenizer=word_tokenize,
                 lower_case=True,
                 max_word_vocab_size=None,
                 min_word_vocab_freq=1,
                 pretrained_word_emb_name="840B",
                 pretrained_word_emb_url=None,
                 target_pretrained_word_emb_name=None,
                 target_pretrained_word_emb_url=None,
                 pretrained_word_emb_cache_dir=".vector_cache/",
                #  pretrained_word_emb_file=None,
                 word_emb_size=None,
                 share_vocab=True):
        super(VocabModel, self).__init__()
        self.tokenizer = tokenizer

        print('Building vocabs...')
        all_words = VocabModel.collect_vocabs(data_set, self.tokenizer, lower_case=lower_case, share_vocab=share_vocab)
        # print('Number of words: {}'.format(len(all_words)))
        if share_vocab:
            in_all_words, out_all_words = all_words, None
            if pretrained_word_emb_name is not None and target_pretrained_word_emb_name is not None:
                warnings.warn("Warning: share vocabulary for source and target language but use different pretrained word embeddings")
        else:
            in_all_words, out_all_words = all_words
            if pretrained_word_emb_name is not None and target_pretrained_word_emb_name is None:
                warnings.warn("Warning: use separate vocabularies for source and target language but same pretrained word embeddings")

        self.in_word_vocab = Vocab(lower_case=lower_case, tokenizer=self.tokenizer)
        self.in_word_vocab.build_vocab(in_all_words, max_vocab_size=max_word_vocab_size, min_vocab_freq=min_word_vocab_freq)

        if pretrained_word_emb_name is not None:
            self.in_word_vocab.load_embeddings(pretrained_word_emb_name=pretrained_word_emb_name, pretrained_word_emb_url=pretrained_word_emb_url, 
                                            pretrained_word_emb_cache_dir=pretrained_word_emb_cache_dir)
            print('Using pretrained word embeddings')
        else:
            self.in_word_vocab.randomize_embeddings(word_emb_size)

        if out_all_words is not None:
            self.out_word_vocab = Vocab(lower_case=lower_case, tokenizer=self.tokenizer)
            self.out_word_vocab.build_vocab(out_all_words, max_vocab_size=max_word_vocab_size, min_vocab_freq=min_word_vocab_freq)

            if target_pretrained_word_emb_name is not None:
                self.out_word_vocab.load_embeddings(pretrained_word_emb_name=target_pretrained_word_emb_name, pretrained_word_emb_url=target_pretrained_word_emb_url, 
                                                pretrained_word_emb_cache_dir=pretrained_word_emb_cache_dir)
                print('Using pretrained word embeddings')
            elif pretrained_word_emb_name is not None:
                self.out_word_vocab.load_embeddings(pretrained_word_emb_name=pretrained_word_emb_name, pretrained_word_emb_url=pretrained_word_emb_url, 
                                                pretrained_word_emb_cache_dir=pretrained_word_emb_cache_dir)
                print('Using pretrained word embeddings')
            else:
                self.out_word_vocab.randomize_embeddings(word_emb_size)
        else:
            self.out_word_vocab = self.in_word_vocab

        if share_vocab:
            print('[ Initialized word embeddings: {} ]'.format(self.in_word_vocab.embeddings.shape))
        else:
            print('[ Using separate word vocabs for input & output text ]')
            print('[ Initialized input word embeddings: {} ]'.format(self.in_word_vocab.embeddings.shape))
            print('[ Initialized output word embeddings: {} ]'.format(self.out_word_vocab.embeddings.shape))

    @classmethod
    def build(cls, saved_vocab_file,
              data_set=None,
              tokenizer=word_tokenize,
              lower_case=True,
              max_word_vocab_size=None,
              min_word_vocab_freq=1,
              pretrained_word_emb_name="840B",
              pretrained_word_emb_url=None,
              target_pretrained_word_emb_name=None,
              target_pretrained_word_emb_url=None,
              pretrained_word_emb_cache_dir=".vector_cache/",
              word_emb_size=None,
              share_vocab=True):
        """Static method for loading a VocabModel from disk.

        Parameters:
        -------
        saved_vocab_file : str
            Path to the saved vocab file.
        data_set: iterable
            A list of instances where each instance is a list of str.
        tokenizer: function, optional
            Word tokenization function, default: nltk.tokenize.word_tokenize.
        max_word_vocab_size: int, optional
            Maximal word vocab size, default: ``None``.
        min_word_vocab_freq: int, optional
            Minimal word vocab frequency, default: ``1``.
        pretrained_word_emb_name: str, optional, default="840B"
            The name of pretrained word embedding in ``torchtext``.
            If it is set ``None``, we will randomly set the initial word embedding values.
        pretrained_word_emb_url: str optional, default: ``None``
            The url for downloading pretrained word embedding.
            Note that we only prepare the default ``url`` for English with ``pretrained_word_emb_name`` as ``"42B"``, ``"840B"``, 'twitter.27B' and '6B'.
        target_pretrained_word_emb_name: str, optional, default=None
            The name of pretrained word embedding in ``torchtext`` for target language.
            If it is set ``None``, we will use ``pretrained_word_emb_name``.
        target_pretrained_word_emb_url: str optional, default: ``None``
            The url for downloading pretrained word embedding for target language.
            Note that we only prepare the default ``url`` for English with ``pretrained_word_emb_name`` as ``"42B"``, ``"840B"``, 'twitter.27B' and '6B'.
        pretrained_word_emb_cache_dir: str, optional, default: ``".vector_cache/"``
            The path of directory saving the temporary word embedding file.
        word_emb_size: int, optional
            Word embedding size, default: ``None``.
        share_vocab : boolean
            Specify whether to share vocab between input and output text, default: ``True``.

        Returns:
        -------
        VocabModel
            Loaded Vocabulary.
        """
        if os.path.exists(saved_vocab_file):
            print('Loading pre-built vocab model stored in {}'.format(saved_vocab_file))
            with open(saved_vocab_file, 'rb') as f:
                vocab_model = pickle.load(f)

        else:
            vocab_model = cls(data_set=data_set, tokenizer=tokenizer,
                              max_word_vocab_size=max_word_vocab_size,
                              min_word_vocab_freq=min_word_vocab_freq,
                              pretrained_word_emb_name=pretrained_word_emb_name,
                              pretrained_word_emb_url=pretrained_word_emb_url,
                              pretrained_word_emb_cache_dir=pretrained_word_emb_cache_dir,
                              target_pretrained_word_emb_name=target_pretrained_word_emb_name,
                              target_pretrained_word_emb_url=target_pretrained_word_emb_url,
                              word_emb_size=word_emb_size,
                              lower_case=lower_case,
                              share_vocab=share_vocab)
            print('Saving vocab model to {}'.format(saved_vocab_file))
            pickle.dump(vocab_model, open(saved_vocab_file, 'wb'))

        return vocab_model

    @staticmethod
    def collect_vocabs(all_instances, tokenizer, lower_case=True, share_vocab=True):
        """Count vocabulary tokens."""
        if share_vocab:
            all_words = Counter()
        else:
            all_words = [Counter(), Counter()]

        for instance in all_instances:
            extracted_tokens = instance.extract()
            if share_vocab:
                all_words.update(extracted_tokens)
            else:
                all_words[0].update(extracted_tokens[0])
                all_words[1].update(extracted_tokens[1])

        return all_words


class WordEmbModel(Vectors):
    def __init__(self, pretrained_word_emb_name="840B", pretrained_word_emb_url=None, pretrained_word_emb_cache_dir=".vector_cache/"):
        
        if pretrained_word_emb_name in GloVe.url.keys() and pretrained_word_emb_url is None:
            url = GloVe.url[pretrained_word_emb_name]
            name = 'glove.{}.{}d.txt'.format(pretrained_word_emb_name, str(300))
        else:
            url = pretrained_word_emb_url
            name = pretrained_word_emb_name
        super().__init__(name=name, cache=pretrained_word_emb_cache_dir, url=url)

    def __getitem__(self, token):
        if token in self.stoi:
            return self.vectors[self.stoi[token]], True
        else:
            return torch.Tensor(np.array(np.random.uniform(low=-0.08, high=0.08, size=(self.dim)),
                                               dtype=np.float)), False

    def get_vecs_by_tokens(self, tokens, lower_case_backup=False):
        """Look up embedding vectors of tokens.

        Arguments:
            tokens: a token or a list of tokens. if `tokens` is a string,
                returns a 1-D tensor of shape `self.dim`; if `tokens` is a
                list of strings, returns a 2-D tensor of shape=(len(tokens),
                self.dim).
            lower_case_backup : Whether to look up the token in the lower case.
                If False, each token in the original case will be looked up;
                if True, each token in the original case will be looked up first,
                if not found in the keys of the property `stoi`, the token in the
                lower case will be looked up. Default: False.

        Examples:
            >>> examples = ['chip', 'baby', 'Beautiful']
            >>> vec = text.vocab.GloVe(name='6B', dim=50)
            >>> ret = vec.get_vecs_by_tokens(tokens, lower_case_backup=True)
        """
        to_reduce = False

        if not isinstance(tokens, list):
            tokens = [tokens]
            to_reduce = True

        hit = 0
        cnt = 0
        indices = []
        if not lower_case_backup:
            for token in tokens:
                result = self[token]
                indices.append(result[0])
                hit += int(result[1])
                cnt += 1
        else:
            for token in tokens:
                result = self[token] if token in self.stoi else self[token.lower()]
                indices.append(result[0])
                hit += int(result[1])
                cnt += 1

        vecs = torch.stack(indices)
        return vecs[0] if to_reduce else vecs, hit, cnt

class Vocab(object):
    """Vocab class.

    Parameters
    ----------
    lower_case : boolean
        Specify whether to lowercase text, default: ``True``.
    tokenizer: function, optional
        Word tokenization function, default: nltk.tokenize.word_tokenize.

    Examples
    -------
    >>> word_vocab = Vocab()
    >>> word_vocab.build_vocab({'i': 10, 'like': 5, 'nlp': 3})
    >>> print(word_vocab.get_vocab_size())
    """
    PAD = 0
    SOS = 1
    EOS = 2
    UNK = 3
    pad_token = constants._PAD_TOKEN
    sos_token = constants._SOS_TOKEN
    eos_token = constants._EOS_TOKEN
    unk_token = constants._UNK_TOKEN
    def __init__(self, lower_case=True, tokenizer=word_tokenize):
        super(Vocab, self).__init__()
        self.lower_case = lower_case
        self.tokenizer = tokenizer
        self.reserved = [self.pad_token, self.sos_token, self.eos_token, self.unk_token]
        self.index2word = self.reserved[:]
        self.word2index = dict(zip(self.reserved, range(len(self.reserved))))
        self.word2count = Counter()
        self.embeddings = None

    def build_vocab(self, vocab_counter, max_vocab_size=None, min_vocab_freq=1):
        """Build vocab from ``vocab_counter`` which is a vocab count dict.

        Parameters
        ----------
        vocab_counter : dict
            Vocab counter.
        max_vocab_size : None, optional
            Maximal word vocab size, default: ``None``.
        min_vocab_freq : int, optional
            Minimal word vocab frequency, default: ``1``.
        """
        self.word2count = vocab_counter
        self._add_words(vocab_counter.keys())
        self._trim(max_vocab_size=max_vocab_size, min_vocab_freq=min_vocab_freq)

    def _add_words(self, words):
        # words: a list of str
        for word in words:
            if self.lower_case:
                word = word.lower()

            if word not in self.word2index:
                self.word2index[word] = len(self.index2word)
                self.index2word.append(word)
        assert len(self.word2index) == len(self.index2word)

    def _trim(self, max_vocab_size=None, min_vocab_freq=1):
        """Trim vocab"""
        if min_vocab_freq <= 1 and (max_vocab_size is None or max_vocab_size >= len(self.word2index)):
            return

        ordered_words = sorted(((c, w) for (w, c) in self.word2count.items()), reverse=True)

        if max_vocab_size:
            ordered_words = ordered_words[:max_vocab_size]

        self.index2word = self.reserved[:]
        self.word2index = dict(zip(self.reserved, range(len(self.reserved))))
        self.word2count = Counter()

        for count, word in ordered_words:
            if count < min_vocab_freq: break
            if word not in self.word2index:
                self.word2index[word] = len(self.index2word)
                self.word2count[word] = count
                self.index2word.append(word)

        assert len(self.word2index) == len(self.index2word)

    def load_embeddings(self, pretrained_word_emb_name="840B", pretrained_word_emb_url=None, pretrained_word_emb_cache_dir=".vector_cache/", dtype=np.float32):
        """Load pretrained word embeddings for initialization"""
        word_model = WordEmbModel(pretrained_word_emb_name=pretrained_word_emb_name, 
                                pretrained_word_emb_url=pretrained_word_emb_url, 
                                pretrained_word_emb_cache_dir=pretrained_word_emb_cache_dir)

        word_list = list(self.word2index.keys())

        word_emb, hit, cnt = word_model.get_vecs_by_tokens(tokens=word_list, lower_case_backup=self.lower_case)
        
        print('Pretrained word embeddings hit ratio: {}'.format(hit / cnt))
        self.embeddings = word_emb.numpy()
        self.embeddings[self.PAD] = np.zeros(word_model.dim)

    def randomize_embeddings(self, n_dims, scale=0.08):
        """Use random word embeddings for initialization."""
        vocab_size = self.get_vocab_size()
        shape = (vocab_size, n_dims)
        self.embeddings = np.array(np.random.uniform(low=-scale, high=scale, size=shape), dtype=np.float32)
        self.embeddings[self.PAD] = np.zeros(n_dims)

    def __getitem__(self, item):
        if type(item) is int:
            return self.index2word[item]
        return self.word2index.get(item, self.UNK)

    def __len__(self):
        return len(self.index2word)

    @lru_cache(maxsize=None)
    def is_word(self, token_id: int) -> bool:
        """Return whether the token at `token_id` is a word; False for punctuations."""
        if token_id < 4: return False
        if token_id >= len(self): return True  # OOV is assumed to be words
        token_str = self.index2word[token_id]
        if not word_detector.search(token_str) or token_str == '<P>':
            return False
        return True

    def get_vocab_size(self):
        return len(self.index2word)

    def getIndex(self, word, use_ie=False):
        # For IE Graph
        # word can be a phrase
        if use_ie:
            if self.lower_case:
                word = word.lower()

            ret = []
            for x in word.replace('_',' ').split(' '):
                if x == '':
                    continue
                ret.append(self.word2index.get(x, self.UNK))

            return ret

        if self.lower_case:
            word = word.lower()

        return self.word2index.get(word, self.UNK)

    def getWord(self, idx):
        return self.index2word[idx] if idx < len(self.index2word) else self.unk_token

    def to_word_sequence(self, seq):
        sentence = []
        for idx in seq:
            word = self.getWord(idx)
            sentence.append(word)
        return ' '.join(sentence)

    def to_index_sequence(self, sentence):
        sentence = sentence.strip()
        if self.lower_case:
            sentence = sentence.lower()

        seq = []
        if self.tokenizer is None:
            for word in sentence.split():
                idx = self.getIndex(word)
                seq.append(idx)
        else:
            for word in self.tokenizer(sentence):
                idx = self.getIndex(word)
                seq.append(idx)
        return seq

    def to_index_sequence_for_list(self, words):
        seq = []
        for word in words:
            if self.lower_case:
                word = word.lower()

            idx = self.getIndex(word)
            seq.append(idx)
        return seq



from torch import nn
import torch as th
import torch.autograd as autograd
from torch.nn.utils.rnn import pad_sequence
from ..base import NodeClassifierBase
from .BiLSTMFeedForwardNNLayer import BiLSTMFeedForwardNNLayer
from .....data.data import *


class BiLSTMFeedForwardNN(NodeClassifierBase):
    """
    Specific class for node classification task.

    ...

    Attributes
    ----------

    input_size : int 
                 the length of input node embeddings
    num_class: int 
               the number of node catrgoriey for classification
    hidden_size: the hidden size of the linear layer

    Methods
    -------

    forward(node_emb)

        Generate the node classification logits.         

    """     
    def __init__(self, input_size, num_class, hidden_size=None, dropout=0):
        
        super(BiLSTMFeedForwardNN, self).__init__()
        
        self.classifier=BiLSTMFeedForwardNNLayer(input_size, num_class, hidden_size,dropout=dropout)
        self.num_class=num_class

    def forward(self, input_graph):
        r"""
        Forward functions to compute the logits tensor for node classification.
    
      
    
        Parameters
        ----------
    
        input graph : GraphData
                     The tensors stored in the node feature field named "node_emb"  in the 
                     input_graph are used  for classification.
                     GraphData are bacthed and needs to unbatch to each sentence.

    
        Returns 
        ---------
        
        output_graph : GraphData
                      The computed logit tensor for each nodes in the graph are stored
                      in the node feature field named "node_logits".
                      logit tensor shape is: [num_class] 
        """ 
        if input_graph.batch_size is None:
           node_emb_padded=input_graph.node_features['node_emb']
           len_emb=node_emb_padded.shape[0]
           bilstm_emb=self.classifier(node_emb_padded.reshape([1,len_emb,-1])) #dimension: batch*max_sent_len*emb_len                     
           input_graph.node_features['logits']=bilstm_emb.reshape(-1,self.num_class)         
        else:                
            node_emb_padded=input_graph.batch_node_features['node_emb']
            bilstm_emb=self.classifier(node_emb_padded) #dimension: batch*max_sent_len*emb_len                     
            input_graph.batch_node_features['logits']=bilstm_emb
            
        return input_graph






import numpy as np


def make_batches(size, batch_size):
    nb_batch = int(np.ceil(size/float(batch_size)))
    return [(i*batch_size, min(size, (i+1)*batch_size)) for i in range(0, nb_batch)] # zgwang: starting point of each batch

def pad_2d_vals_no_size(in_vals, dtype=np.int32):
    size1 = len(in_vals)
    size2 = np.max([len(x) for x in in_vals])
    return pad_2d_vals(in_vals, size1, size2, dtype=dtype)

def pad_2d_vals(in_vals, dim1_size, dim2_size, dtype=np.int32):
    out_val = np.zeros((dim1_size, dim2_size), dtype=dtype)
    if dim1_size > len(in_vals): dim1_size = len(in_vals)
    for i in range(dim1_size):
        cur_in_vals = in_vals[i]
        cur_dim2_size = dim2_size
        if cur_dim2_size > len(cur_in_vals): cur_dim2_size = len(cur_in_vals)
        out_val[i,:cur_dim2_size] = cur_in_vals[:cur_dim2_size]
    return out_val

def pad_3d_vals_no_size(in_vals, dtype=np.int32):
    size1 = len(in_vals)
    size2 = np.max([len(x) for x in in_vals])
    size3 = 0
    for val in in_vals:
        cur_size3 = np.max([len(x) for x in val])
        if size3<cur_size3: size3 = cur_size3
    return pad_3d_vals(in_vals, size1, size2, size3, dtype=dtype)

def pad_3d_vals(in_vals, dim1_size, dim2_size, dim3_size, dtype=np.int32):
#     print(in_vals)
    out_val = np.zeros((dim1_size, dim2_size, dim3_size), dtype=dtype)
    if dim1_size > len(in_vals): dim1_size = len(in_vals)
    for i in range(dim1_size):
        in_vals_i = in_vals[i]
        cur_dim2_size = dim2_size
        if cur_dim2_size > len(in_vals_i): cur_dim2_size = len(in_vals_i)
        for j in range(cur_dim2_size):
            in_vals_ij = in_vals_i[j]
            cur_dim3_size = dim3_size
            if cur_dim3_size > len(in_vals_ij): cur_dim3_size = len(in_vals_ij)
            out_val[i, j, :cur_dim3_size] = in_vals_ij[:cur_dim3_size]
    return out_val

def pad_4d_vals(in_vals, dim1_size, dim2_size, dim3_size, dim4_size, dtype=np.int32):
    out_val = np.zeros((dim1_size, dim2_size, dim3_size, dim4_size), dtype=dtype)
    if dim1_size > len(in_vals): dim1_size = len(in_vals)
    for i in range(dim1_size):
        in_vals_i = in_vals[i]
        cur_dim2_size = dim2_size
        if cur_dim2_size > len(in_vals_i): cur_dim2_size = len(in_vals_i)
        for j in range(cur_dim2_size):
            in_vals_ij = in_vals_i[j]
            cur_dim3_size = dim3_size
            if cur_dim3_size > len(in_vals_ij): cur_dim3_size = len(in_vals_ij)
            for k in range(cur_dim3_size):
                in_vals_ijk = in_vals_ij[k]
                cur_dim4_size = dim4_size
                if cur_dim4_size > len(in_vals_ijk): cur_dim4_size = len(in_vals_ijk)
                out_val[i, j, k, :cur_dim4_size] = in_vals_ijk[:cur_dim4_size]
    return out_val

def pad_target_labels(in_val, max_length, dtype=np.float32):
    batch_size = len(in_val)
    out_val = np.zeros((batch_size, max_length), dtype=dtype)
    for i in range(batch_size):
        for index in in_val[i]:
            out_val[i,index] = 1.0
    return out_val



import json

from stanfordcorenlp import StanfordCoreNLP

from graph4nlp.pytorch.data.data import GraphData, to_batch
from graph4nlp.pytorch.modules.utils.vocab_utils import VocabModel
from graph4nlp.pytorch.modules.graph_construction.base import StaticGraphConstructionBase
import dgl
import networkx as nx
from graph4nlp.pytorch.modules.utils.padding_utils import pad_2d_vals
import numpy as np
import torch


class IEBasedGraphConstruction(StaticGraphConstructionBase):
    """
        Information Extraction based graph construction class

    Parameters
    ----------
    embedding_style: dict
        Specify embedding styles including ``single_token_item``, ``emb_strategy``, ``num_rnn_layers``, ``bert_model_name`` and ``bert_lower_case``.
    vocab: VocabModel
        Vocabulary including all words appeared in graphs.
    """

    def __init__(self, embedding_style, vocab, hidden_size=300, fix_word_emb=True, fix_bert_emb=True, word_dropout=None,
                 rnn_dropout=None, device=None):
        super(IEBasedGraphConstruction, self).__init__(word_vocab=vocab,
                                                       embedding_styles=embedding_style,
                                                       hidden_size=hidden_size,
                                                       fix_word_emb=fix_word_emb,
                                                       fix_bert_emb=fix_bert_emb,
                                                       word_dropout=word_dropout,
                                                       rnn_dropout=rnn_dropout)
        self.vocab = vocab
        self.verbase = 1

    def add_vocab(self, g):
        """
            Add node tokens appeared in graph g to vocabulary.

        Parameters
        ----------
        g: GraphData
            Graph data-structure.

        """
        for i in range(g.get_node_num()):
            attr = g.get_node_attrs(i)[i]
            self.vocab.word_vocab._add_words([attr["token"]])

    @classmethod
    def parsing(cls, all_sent_triples_list, edge_strategy):
        """
        Parameters
        ----------
        all_sent_triples_list: list
        edge_strategy: str

        Returns
        -------
        parsed_results: dict
            parsed_results is an intermediate dict that contains all the information of
            the constructed IE graph for a piece of raw text input.

            `parsed_results['graph_content']` is a list of dict.

            Each dict in `parsed_results['graph_content']` contains information about a
            triple (src_ent, rel, tgt_ent).

            `parsed_results['graph_nodes']` contains all nodes in the KG graph.

            `parsed_results['node_num']` is the number of nodes in the KG graph.
        """

        parsed_results = {}
        parsed_results['graph_content'] = []
        graph_nodes = []
        for triple in all_sent_triples_list:
            if edge_strategy is None:
                if triple[0] not in graph_nodes:
                    graph_nodes.append(triple[0])

                if triple[2] not in graph_nodes:
                    graph_nodes.append(triple[2])

                triple_info = {'edge_tokens': triple[1],
                               'src': {
                                   'tokens': triple[0],
                                   'id': graph_nodes.index(triple[0])
                               },
                               'tgt': {
                                   'tokens': triple[2],
                                   'id': graph_nodes.index(triple[2])
                               }}
                if triple_info not in parsed_results['graph_content']:
                    parsed_results['graph_content'].append(triple_info)
            elif edge_strategy == "as_node":
                if triple[0] not in graph_nodes:
                    graph_nodes.append(triple[0])

                if triple[1] not in graph_nodes:
                    graph_nodes.append(triple[1])

                if triple[2] not in graph_nodes:
                    graph_nodes.append(triple[2])

                triple_info_0_1 = {'edge_tokens': [],
                                   'src': {
                                       'tokens': triple[0],
                                       'id': graph_nodes.index(triple[0]),
                                       'type': 0,  # 'ent_node'
                                   },
                                   'tgt': {
                                       'tokens': triple[1],
                                       'id': graph_nodes.index(triple[1]),
                                       'type': 3,  # 'edge_node'
                                   }}

                triple_info_1_2 = {'edge_tokens': [],
                                   'src': {
                                       'tokens': triple[1],
                                       'id': graph_nodes.index(triple[1]),
                                       'type': 3,  # 'edge_node'
                                   },
                                   'tgt': {
                                       'tokens': triple[2],
                                       'id': graph_nodes.index(triple[2]),
                                       'type': 0,  # 'ent_node'
                                   }}

                if triple_info_0_1 not in parsed_results['graph_content']:
                    parsed_results['graph_content'].append(triple_info_0_1)
                if triple_info_1_2 not in parsed_results['graph_content']:
                    parsed_results['graph_content'].append(triple_info_1_2)
            else:
                raise NotImplementedError()

        parsed_results['node_num'] = len(graph_nodes)
        parsed_results['graph_nodes'] = graph_nodes

        return parsed_results

    @classmethod
    def topology(cls, raw_text_data, nlp_processor, processor_args, merge_strategy, edge_strategy, verbase=True):
        """
            Graph building method.

        Parameters
        ----------
        raw_text_data: str
            Raw text data, it can be multi-sentences.

        nlp_processor: StanfordCoreNLP
            NLP parsing tools

        merge_strategy: None or str, option=[None, "global", "user_define"]
            Strategy to merge sub-graphs into one graph
            ``None``:  Do not add additional nodes and edges.

            ``global``: All subjects in extracted triples are connected by a "GLOBAL_NODE"
                        using a "global" edge

            ``"user_define"``: We will give this option to the user. User can override this method to define your merge
                               strategy.

        edge_strategy: None or str, option=[None, "as_node"]
            Strategy to process edge.
            ``None``: It will be the default option.
                      Edge information will be preserved in GraphDate.edge_attributes.
            ``as_node``: We will view the edge as a graph node.
                         If there is an edge whose type is ``k`` between node ``i`` and node ``j``,
                         we will insert a node ``k`` into the graph and link node (``i``, ``k``) and (``k``, ``j``).
                         The ``type`` of original nodes will be set as ``ent_node``,
                         while the ``type`` of edge nodes is ``edge_node`.`

        Returns
        -------
        graph: GraphData
            The merged graph data-structure.
        """
        cls.verbase = verbase

        if isinstance(processor_args, list):
            props_coref = processor_args[0]
            props_openie = processor_args[1]
        else:
            raise RuntimeError('processor_args for IEBasedGraphConstruction shouble be a list of dict.')

        # Do coreference resolution on the whole 'raw_text_data'
        coref_json = nlp_processor.annotate(raw_text_data.strip(), properties=props_coref)
        from .utils import CORENLP_TIMEOUT_SIGNATURE
        if CORENLP_TIMEOUT_SIGNATURE in coref_json:
            raise TimeoutError('Coref-CoreNLP timed out at input: \n{}\n This item will be skipped. '
                               'Please check the input or change the timeout threshold.'.format(raw_text_data))

        coref_dict = json.loads(coref_json)

        # Extract and preserve necessary parsing results from coref_dict['sentences']
        # sent_dict['tokenWords']: list of tokens in a sentence
        sentences = []
        for sent in coref_dict['sentences']:
            sent_dict = {}
            sent_dict['sentNum'] = sent['index']  # start from 0
            sent_dict['tokens'] = sent['tokens']
            sent_dict['tokenWords'] = [token['word'] for token in sent['tokens']]
            sent_dict['sentText'] = ' '.join(sent_dict['tokenWords'])
            sentences.append(sent_dict)

        for k, v in coref_dict['corefs'].items():
            # v is a list of dict, each dict contains a str
            # v[0] contains 'original entity str'
            # v[1:] contain 'pron strs' refers to 'original entity str'
            ent_text = v[0]['text']  # 'original entity str'
            if ',' in ent_text:
                # cut the 'original entity str' if it is too long
                ent_text = ent_text.split(',')[0].strip()
            ent_sentNum = v[0]['sentNum'] - 1  # the sentNum 'original entity str' appears in
            ent_startIndex = v[0]['startIndex'] - 1  # the startIndex 'original entity str' appears in
            ent_endIndex = v[0]['endIndex'] - 1  # the endIndex 'original entity str' appears in

            for pron in v[1:]:
                pron_text = pron['text']  # 'pron strs'
                if ent_text == pron_text or v[0]['text'] == pron_text:
                    continue
                pron_sentNum = pron['sentNum'] - 1  # the sentNum 'pron str' appears in
                pron_startIndex = pron['startIndex'] - 1
                pron_endIndex = pron['endIndex'] - 1

                # replace 'pron str' with 'original entity str'
                sentences[pron_sentNum]['tokenWords'][pron_startIndex] = ent_text
                for rm_idx in range(pron_startIndex+1, pron_endIndex):
                    sentences[pron_sentNum]['tokenWords'][rm_idx] = ""

        # build resolved text
        for sent_id, _ in enumerate(sentences):
            sentences[sent_id]['tokenWords'] = list(filter(lambda a: a != "", sentences[sent_id]['tokenWords']))
            sentences[sent_id]['resolvedText'] = ' '.join(sentences[sent_id]['tokenWords'])

        # use OpenIE to extract triples from resolvedText
        all_sent_triples = {}
        for sent in sentences:
            resolved_sent = sent['resolvedText']
            openie_json = nlp_processor.annotate(resolved_sent.strip(), properties=props_openie)
            if CORENLP_TIMEOUT_SIGNATURE in openie_json:
                raise TimeoutError('OpenIE-CoreNLP timed out at input: \n{}\n This item will be skipped. '
                                   'Please check the input or change the timeout threshold.'.format(raw_text_data))
            openie_dict = json.loads(openie_json)
            for triple_dict in openie_dict['sentences'][0]['openie']:
                sbj = triple_dict['subject']
                rel = triple_dict['relation']
                if rel in ['was', 'is', 'were', 'are']:
                    continue
                obj = triple_dict['object']

                # If two triples have the same subject and relation,
                # only preserve the one has longer object
                if sbj+'<TSEP>'+rel not in all_sent_triples.keys():
                    all_sent_triples[sbj+'<TSEP>'+rel] = [sbj, rel, obj]
                else:
                    if len(obj)>len(all_sent_triples[sbj+'<TSEP>'+rel][2]):
                        all_sent_triples[sbj + '<TSEP>' + rel] = [sbj, rel, obj]

        all_sent_triples_list = list(all_sent_triples.values())  # triples extracted from all sentences

        # remove similar triples
        triples_rm_list = []
        for i, lst_i in enumerate(all_sent_triples_list[:-1]):
            for j, lst_j in enumerate(all_sent_triples_list[i+1:]):
                str_i = ' '.join(lst_i)
                str_j = ' '.join(lst_j)
                if str_i in str_j or str_j in str_i or \
                        lst_i[0]+lst_i[2]==lst_j[0]+lst_j[2] or \
                        lst_i[1]+lst_i[2]==lst_j[1]+lst_j[2]:
                    if len(lst_i[1])>len(lst_j[1]):
                        triples_rm_list.append(lst_j)
                    else:
                        triples_rm_list.append(lst_i)

        for lst in triples_rm_list:
            if lst in all_sent_triples_list:
                all_sent_triples_list.remove(lst)

        global_triples = cls._graph_connect(all_sent_triples_list, merge_strategy)
        all_sent_triples_list.extend(global_triples)

        parsed_results = cls.parsing(all_sent_triples_list, edge_strategy)

        graph = cls._construct_static_graph(parsed_results, edge_strategy=edge_strategy)

        if cls.verbase:
            for info in parsed_results['graph_content']:
                print(info)
            try:
                print("is_connected="+str(nx.is_connected(nx.Graph(graph.to_dgl().to_networkx()))))
            except:
                print("is_connected=False")

        return graph


    def embedding(self, graph: GraphData):
        node_attributes = graph.node_attributes
        edge_attributes = graph.edge_attributes

        # Build embedding(initial feature vector) for graph nodes.
        # Each node may contains multiple tokens.
        node_idxs_list = []
        node_len_list = []
        for node_id, node_dict in node_attributes.items():
            node_word_idxs = []
            for token in node_dict['token'].split():
                node_word_idxs.append(self.vocab.getIndex(token))
            node_idxs_list.append(node_word_idxs)
            node_len_list.append(len(node_word_idxs))
        max_size = max(node_len_list)
        node_idxs_list = [x+[self.vocab.PAD]*(max_size-len(x)) for x in node_idxs_list]
        node_idxs_tensor = torch.LongTensor(node_idxs_list)
        # if self.embedding_layer.node_edge_emb_strategy == 'mean':
        #     node_len_tensor = torch.LongTensor(node_len_list).view(-1, 1)
        # else:
        node_len_tensor = torch.LongTensor(node_len_list)
        num_nodes = torch.LongTensor([len(node_len_list)])
        node_feat = self.embedding_layer(node_idxs_tensor, node_len_tensor, num_nodes)
        graph.node_features['node_feat'] = node_feat

        if 'token' in edge_attributes[0].keys():
            # If edge information is stored in `edge_attributes`,
            # build embedding(initial feature vector) for graph edges.
            # Each edge may contains multiple tokens.
            edge_idxs_list = []
            edge_len_list = []
            for edge_id, edge_dict in edge_attributes.items():
                edge_word_idxs = []
                for token in edge_dict['token']:
                    edge_word_idxs.append(self.vocab.getIndex(token))
                edge_idxs_list.append(edge_word_idxs)
                edge_len_list.append(len(edge_word_idxs))

            max_size = max(edge_len_list)
            edge_idxs_list = [x + [self.vocab.PAD] * (max_size - len(x)) for x in edge_idxs_list]
            edge_idxs_tensor = torch.LongTensor(edge_idxs_list)
            # if self.embedding_layer.node_edge_emb_strategy == 'mean':
            #     edge_len_tensor = torch.LongTensor(edge_len_list).view(-1, 1)
            # else:
            edge_len_tensor = torch.LongTensor(edge_len_list)
            num_edges = torch.LongTensor([len(edge_len_list)])
            edge_feat = self.embedding_layer(edge_idxs_tensor, edge_len_tensor, num_edges)
            graph.edge_features['edge_feat'] = edge_feat

        return graph


    @classmethod
    def _construct_static_graph(cls, parsed_object, edge_strategy=None):
        """

        Parameters
        ----------
        parsed_object: dict
            ``parsed_object`` contains all triples extracted from the raw_text_data.

        edge_strategy: None or str, option=[None, "as_node"]
            Strategy to process edge.
            ``None``: It will be the default option.
                      Edge information will be preserved in GraphDate.edge_attributes.
            ``as_node``: We will view the edge as a graph node.
                         If there is an edge whose type is ``k`` between node ``i`` and node ``j``,
                         we will insert a node ``k`` into the graph and link node (``i``, ``k``) and (``k``, ``j``).
                         The ``type`` of original nodes will be set as ``ent_node``,
                         while the ``type`` of edge nodes is ``edge_node`.`

        Returns
        -------
        graph: GraphData
            graph structure for single sentence
        """
        ret_graph = GraphData()
        node_num = parsed_object["node_num"]
        if node_num == 0:
            raise RuntimeError('"The number of nodes to be added should be greater than 0. (Got {})"'.format(node_num))
        ret_graph.add_nodes(node_num)
        for triple_info in parsed_object["graph_content"]:
            if edge_strategy is None:
                ret_graph.add_edge(triple_info["src"]['id'], triple_info['tgt']['id'])
                eids = ret_graph.edge_ids(triple_info["src"]['id'], triple_info['tgt']['id'])
                for eid in eids:
                    ret_graph.edge_attributes[eid]['token'] = triple_info['edge_tokens']
            elif edge_strategy == 'as_node':
                ret_graph.add_edge(triple_info["src"]['id'], triple_info['tgt']['id'])
            else:
                raise NotImplementedError()

            ret_graph.node_attributes[triple_info["src"]['id']]['token'] = triple_info["src"]['tokens']
            ret_graph.node_attributes[triple_info["tgt"]['id']]['token'] = triple_info['tgt']['tokens']
            if edge_strategy == 'as_node':
                ret_graph.node_attributes[triple_info["src"]['id']]['type'] = triple_info["src"]['type']
                ret_graph.node_attributes[triple_info["tgt"]['id']]['type'] = triple_info['tgt']['type']

        return ret_graph

    @classmethod
    def _graph_connect(cls, triple_list, merge_strategy=None):
        """
            This method will connect entities in the ``triple_list`` to ensure the graph
            is connected.

        Parameters
        ----------
        triple_list: list of [subject, relation, object]
            A list of all triples extracted from ``raw_text_data`` using coref and openie.

        merge_strategy: None or str, option=[None, "global", "user_define"]
            Strategy to merge sub-graphs into one graph
            ``None``:  Do not add additional nodes and edges.

            ``global``: All subjects in extracted triples are connected by a "GLOBAL_NODE"
                        using a "global" edge

            ``"user_define"``: We will give this option to the user. User can override this method to define your merge
                               strategy.

        Returns
        -------
        global_triples: list of [subject, relation, object]
            The added triples using merge_strategy.
        """

        if merge_strategy == 'global':
            graph_nodes = []
            global_triples = []
            for triple in triple_list:
                if triple[0] not in graph_nodes:
                    graph_nodes.append(triple[0])
                    global_triples.append([triple[0], 'global', 'GLOBAL_NODE'])

                if triple[2] not in graph_nodes:
                    graph_nodes.append(triple[2])

            return global_triples
        elif merge_strategy == None:
            return []
        else:
            raise NotImplementedError()

    def forward(self, batch_graphdata: list):
        batch_graphdata = self.embedding_layer(batch_graphdata)
        return batch_graphdata



# Views implementations used in GraphData
from collections import namedtuple

from .utils import slice_to_list

NodeRepr = namedtuple('NodeData', ['features', 'attributes'])
EdgeRepr = namedtuple('EdgeData', ['features'])


class NodeView(object):
    """
    View for graph nodes at at high level.
    """

    def __init__(self, graph):
        self._graph = graph

    def __getitem__(self, node_idx: int or slice) -> NodeRepr:
        """
        Get a number of nodes and their corresponding data.

        Parameters
        ----------
        node_idx: int or slice
            The index of nodes to be accessed.

        Returns
        -------
        NodeRepr
            A collection of the corresponding data.
        """
        # consistency check
        # 1. type check
        assert isinstance(node_idx, int) or isinstance(node_idx, slice), "Only int and slice are supported currently."
        # 2. boundary check
        if isinstance(node_idx, slice):
            node_idx_list = slice_to_list(node_idx, self._graph.get_node_num())
            for idx in node_idx_list:
                assert idx < self._graph.get_node_num(), 'Node {} does not exist in the graph.'.format(idx)
        else:
            assert node_idx < self._graph.get_node_num(), 'Node {} does not exist in the graph.'.format(node_idx)

        return NodeRepr(features=NodeFeatView(self._graph, node_idx), attributes=NodeAttrView(self._graph, node_idx))

    def __len__(self):
        return self._graph.get_node_num()

    def __call__(self):
        return list(range(self._graph.get_node_num))


class NodeFeatView(object):
    """
    View for node features which are all tensors.
    """

    def __init__(self, graph, nodes: slice or int):
        self._graph = graph
        self._nodes = nodes

    def __getitem__(self, feature_name):
        return self._graph.get_node_features(self._nodes)[feature_name]

    def __setitem__(self, feature_name, feature_value):
        return self._graph.set_node_features(self._nodes, {feature_name: feature_value})

    def __repr__(self):
        return repr(self._graph.get_node_features(self._nodes))

    def keys(self):
        return self._graph.node_feature_names()


class NodeAttrView(object):
    """
    View for node attributes which are arbitrary objects.
    """

    def __init__(self, graph, nodes):
        self._graph = graph
        self._nodes = nodes

    def __getitem__(self, item):
        total_attrs = self._graph.get_node_attrs(self._nodes)
        filtered_attrs = dict()
        for k, v in total_attrs.items():
            if item in v:
                filtered_attrs[k] = v[item]
        return filtered_attrs

    def __setitem__(self, key, value):
        raise NotImplementedError('NodeAttrView does not support modifying node attributes.'
                                  'To modify node attributes, please use GraphData.node_attributes')

    def __repr__(self):
        return repr(self._graph.get_node_attrs(self._nodes))


class EdgeView(object):
    """
    View for edges at high level.
    """

    def __init__(self, graph):
        self._graph = graph

    def __call__(self, *args, **kwargs):
        return self._graph.get_all_edges(*args, **kwargs)

    def __getitem__(self, item):
        return EdgeRepr(features=EdgeFeatView(self._graph, item))

    def __repr__(self):
        return repr({'Edges': self._graph.get_all_edges()})


class EdgeFeatView(object):
    def __init__(self, graph, edges):
        self._graph = graph
        self._edges = edges

    def __getitem__(self, item):
        return self._graph.get_edge_feature(self._edges)[item]

    def __setitem__(self, key, value):
        self._graph.set_edge_feature(self._edges, {key: value})

    def keys(self):
        return self._graph.get_edge_feature_names()


class BatchNodeFeatView(object):
    def __init__(self, graph):
        self._graph = graph

    def __getitem__(self, item):
        return self._graph._get_batch_node_features(item)

    def __setitem__(self, key, value):
        return self._graph._set_batch_node_features(key, value)

    def __repr__(self):
        return self._graph._get_batch_node_features()


class BatchEdgeFeatView(object):
    def __init__(self, graph):
        self._graph = graph

    def __getitem__(self, item):
        return self._graph._get_batch_edge_features(item)

    def __setitem__(self, key, value):
        return self._graph._set_batch_edge_features(key, value)

    def __repr__(self):
        return self._graph._get_batch_edge_features()



import yaml


def update_values(to_args: dict, from_args_list: [dict]):
    """
        update_values(template, [args1, args2])
    Parameters
    ----------
    to_args
    from_args_list

    Returns
    -------

    """
    for from_args in from_args_list:
        if not isinstance(from_args, dict):
            raise TypeError("The element in ``from_args_list`` should be dict")
        update_values_api(dict_from=from_args, dict_to=to_args)


def update_values_api(dict_from, dict_to):
    for key, value in dict_from.items():
        if isinstance(value, dict) and key in dict_to.keys():
            update_values_api(dict_from[key], dict_to[key])
        else:
            dict_to[key] = dict_from[key]


def get_yaml_config(config_path="config.yml"):
    with open(config_path, "r") as setting:
        config = yaml.load(setting)
    return config



import abc

import torch.nn as nn
import torch.nn.functional as F

from graph4nlp.pytorch.modules.graph_construction.constituency_graph_construction import \
    ConstituencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.ie_graph_construction import IEBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.node_embedding_based_graph_construction import \
    NodeEmbeddingBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.node_embedding_based_refined_graph_construction import \
    NodeEmbeddingBasedRefinedGraphConstruction
from graph4nlp.pytorch.modules.graph_embedding.gat import GAT
from graph4nlp.pytorch.modules.graph_embedding.gcn import GCN
from graph4nlp.pytorch.modules.graph_embedding.ggnn import GGNN
from graph4nlp.pytorch.modules.graph_embedding.graphsage import GraphSAGE


class Graph2XBase(nn.Module):
    def __init__(self, vocab_model, embedding_style, graph_type, emb_input_size, emb_hidden_size,

                 gnn, gnn_num_layers, gnn_direction_option, gnn_input_size, gnn_hidden_size, gnn_output_size,

                 # dropout
                 emb_word_dropout=0.0, emb_rnn_dropout=0.0, emb_fix_word_emb=False, emb_fix_bert_emb=False,

                 gnn_feats_dropout=0.0, gnn_attn_dropout=0.0,
                 **kwargs):
        super(Graph2XBase, self).__init__()
        self._build_embedding_encoder(graph_type=graph_type, embedding_style=embedding_style, vocab_model=vocab_model,
                                      emb_input_size=emb_input_size, emb_hidden_size=emb_hidden_size, emb_word_dropout=emb_word_dropout,
                                      emb_rnn_dropout=emb_rnn_dropout, emb_fix_word_emb=emb_fix_word_emb,
                                      emb_fix_bert_emb=emb_fix_bert_emb, **kwargs)

        self._build_gnn_encoder(gnn=gnn, num_layers=gnn_num_layers,
                                input_size=gnn_input_size, hidden_size=gnn_hidden_size, output_size=gnn_output_size,
                                direction_option=gnn_direction_option, feats_dropout=gnn_feats_dropout,
                                gnn_attn_dropout=gnn_attn_dropout, **kwargs)

    def _build_embedding_encoder(self, graph_type, embedding_style, vocab_model,
                                 emb_input_size, emb_hidden_size, emb_rnn_dropout,
                                 emb_word_dropout,
                                 # dynamic parameters
                                 emb_sim_metric_type=None, emb_num_heads=None, emb_top_k_neigh=None,
                                 emb_epsilon_neigh=None,
                                 emb_smoothness_ratio=None, emb_connectivity_ratio=None, emb_sparsity_ratio=None,
                                 emb_alpha_fusion=None,
                                 emb_fix_word_emb=False, emb_fix_bert_emb=False, **kwargs):

        if not isinstance(graph_type, str):
            raise ValueError("graph_type parameter should be str")

        if graph_type == "dependency":
            self.graph_topology = DependencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                   vocab=vocab_model.in_word_vocab,
                                                                   hidden_size=emb_hidden_size,
                                                                   word_dropout=emb_word_dropout,
                                                                   rnn_dropout=emb_rnn_dropout,
                                                                   fix_word_emb=emb_fix_word_emb,
                                                                   fix_bert_emb=emb_fix_bert_emb)
        elif graph_type == "constituency":
            self.graph_topology = ConstituencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                     vocab=vocab_model.in_word_vocab,
                                                                     hidden_size=emb_hidden_size,
                                                                     word_dropout=emb_word_dropout,
                                                                     rnn_dropout=emb_rnn_dropout,
                                                                     fix_word_emb=emb_fix_word_emb)
        elif graph_type == "ie":
            self.graph_topology = IEBasedGraphConstruction(embedding_style=embedding_style,
                                                           vocab=vocab_model.in_word_vocab,
                                                           hidden_size=emb_hidden_size,
                                                           word_dropout=emb_word_dropout,
                                                           rnn_dropout=emb_rnn_dropout,
                                                           fix_word_emb=emb_fix_word_emb)
        elif graph_type == "node_emb":
            self.graph_topology = NodeEmbeddingBasedGraphConstruction(
                vocab_model.in_word_vocab,
                embedding_style,
                sim_metric_type=emb_sim_metric_type,
                num_heads=emb_num_heads,
                top_k_neigh=emb_top_k_neigh,
                epsilon_neigh=emb_epsilon_neigh,
                smoothness_ratio=emb_smoothness_ratio,
                connectivity_ratio=emb_connectivity_ratio,
                sparsity_ratio=emb_sparsity_ratio,
                input_size=emb_input_size,
                hidden_size=emb_hidden_size,
                fix_word_emb=emb_fix_word_emb,
                fix_bert_emb=emb_fix_bert_emb,
                word_dropout=emb_word_dropout,
                rnn_dropout=emb_rnn_dropout)
        elif graph_type == "node_emb_refined":
            self.graph_topology = NodeEmbeddingBasedRefinedGraphConstruction(
                vocab_model.in_word_vocab,
                embedding_style,
                emb_alpha_fusion,
                sim_metric_type=emb_sim_metric_type,
                num_heads=emb_num_heads,
                top_k_neigh=emb_top_k_neigh,
                epsilon_neigh=emb_epsilon_neigh,
                smoothness_ratio=emb_smoothness_ratio,
                connectivity_ratio=emb_connectivity_ratio,
                sparsity_ratio=emb_sparsity_ratio,
                input_size=emb_input_size,
                hidden_size=emb_hidden_size,
                fix_word_emb=emb_fix_word_emb,
                fix_bert_emb=emb_fix_bert_emb,
                word_dropout=emb_word_dropout,
                rnn_dropout=emb_rnn_dropout)
        else:
            raise NotImplementedError()
        self.enc_word_emb = self.graph_topology.embedding_layer.word_emb_layers['w2v'] if 'w2v' in self.graph_topology.embedding_layer.word_emb_layers else None

    def _build_gnn_encoder(self, gnn, num_layers, input_size, hidden_size, output_size, direction_option, feats_dropout,
                           gnn_heads=None, gnn_use_residual=True, gnn_attn_dropout=0.0, gnn_activation=F.relu,  # gat
                           gnn_bias=True, gnn_allow_zero_in_degree=True, gnn_norm='both', gnn_weight=True,
                           gnn_use_edge_weight=False, gnn_gcn_norm='both',  # gcn
                           gnn_n_etypes=1,  # ggnn
                           gnn_aggregator_type="lstm",  # graphsage
                           **kwargs):
        if gnn == "gat":
            self.gnn_encoder = GAT(num_layers, input_size, hidden_size, output_size, gnn_heads,
                                   direction_option=direction_option,
                                   feat_drop=feats_dropout, attn_drop=gnn_attn_dropout, activation=gnn_activation,
                                   residual=gnn_use_residual)
        elif gnn == "ggnn":
            self.gnn_encoder = GGNN(num_layers, input_size, hidden_size, output_size, direction_option=direction_option,
                                    use_edge_weight=gnn_use_edge_weight, feat_drop=feats_dropout, n_etypes=gnn_n_etypes)
        elif gnn == "graphsage":
            self.gnn_encoder = GraphSAGE(num_layers, input_size, hidden_size, output_size,
                                         aggregator_type=gnn_aggregator_type,
                                         direction_option=direction_option, feat_drop=feats_dropout,
                                         activation=gnn_activation, bias=gnn_bias, use_edge_weight=gnn_use_edge_weight)
        elif gnn == "gcn":
            self.gnn_encoder = GCN(num_layers, input_size, hidden_size, output_size,
                                   direction_option=direction_option, weight=gnn_weight, gcn_norm=gnn_gcn_norm,
                                   allow_zero_in_degree=gnn_allow_zero_in_degree, activation=gnn_activation,
                                   use_edge_weight=gnn_use_edge_weight)
        else:
            raise NotImplementedError()

    @abc.abstractmethod
    def __build_decoder(self):
        raise NotImplementedError()



#!/usr/bin/env python
# 
# File Name : rouge_tool.py
#
# Description : Computes ROUGE-L metric as described by Lin and Hovey (2004)
#
# Creation Date : 2015-01-07 06:03
# Author : Ramakrishna Vedantam <vrama91@vt.edu>

import numpy as np
import pdb

def my_lcs(string, sub):
    """
    Calculates longest common subsequence for a pair of tokenized strings
    :param string : list of str : tokens from a string split using whitespace
    :param sub : list of str : shorter string, also split using whitespace
    :returns: length (list of int): length of the longest common subsequence between the two strings

    Note: my_lcs only gives length of the longest common subsequence, not the actual LCS
    """
    if(len(string)< len(sub)):
        sub, string = string, sub

    lengths = [[0 for i in range(0,len(sub)+1)] for j in range(0,len(string)+1)]

    for j in range(1,len(sub)+1):
        for i in range(1,len(string)+1):
            if(string[i-1] == sub[j-1]):
                lengths[i][j] = lengths[i-1][j-1] + 1
            else:
                lengths[i][j] = max(lengths[i-1][j] , lengths[i][j-1])

    return lengths[len(string)][len(sub)]

class Rouge():
    '''
    Class for computing ROUGE-L score for a set of candidate sentences for the MS COCO test set

    '''
    def __init__(self):
        # vrama91: updated the value below based on discussion with Hovey
        self.beta = 1.2

    def calc_score(self, candidate, refs):
        """
        Compute ROUGE-L score given one candidate and references for an image
        :param candidate: str : candidate sentence to be evaluated
        :param refs: list of str : COCO reference sentences for the particular image to be evaluated
        :returns score: int (ROUGE-L score for the candidate evaluated against references)
        """
        assert(len(candidate)==1)	
        assert(len(refs)>0)         
        prec = []
        rec = []

        # split into tokens
        token_c = candidate[0].split(" ")
    	
        for reference in refs:
            # split into tokens
            token_r = reference.split(" ")
            # compute the longest common subsequence
            lcs = my_lcs(token_r, token_c)
            prec.append(lcs/float(len(token_c)))
            rec.append(lcs/float(len(token_r)))

        prec_max = max(prec)
        rec_max = max(rec)

        if(prec_max!=0 and rec_max !=0):
            score = ((1 + self.beta**2)*prec_max*rec_max)/float(rec_max + self.beta**2*prec_max)
        else:
            score = 0.0
        return score

    def compute_score(self, gts, res):
        """
        Computes Rouge-L score given a set of reference and candidate sentences for the dataset
        Invoked by evaluate_captions.py 
        :param hypo_for_image: dict : candidate / test sentences with "image name" key and "tokenized sentences" as values 
        :param ref_for_image: dict : reference MS-COCO sentences with "image name" key and "tokenized sentences" as values
        :returns: average_score: float (mean ROUGE-L score computed by averaging scores for all the images)
        """
        assert(gts.keys() == res.keys())
        imgIds = gts.keys()

        score = []
        for id in imgIds:
            hypo = res[id]
            ref  = gts[id]

            score.append(self.calc_score(hypo, ref))

            # Sanity check.
            assert(type(hypo) is list)
            assert(len(hypo) == 1)
            assert(type(ref) is list)
            assert(len(ref) > 0)

        average_score = np.mean(np.array(score))
        return average_score, np.array(score)

    def method(self):
        return "Rouge"



from .general_loss import GeneralLoss
from .coverage_loss import CoverageLoss
from .seq_generation_loss import SeqGenerationLoss
from .kg_loss import KGLoss

__all__ = ["GeneralLoss", "CoverageLoss", "SeqGenerationLoss", "KGLoss"]



import random
from functools import reduce

import torch
import torch.nn as nn

from graph4nlp.pytorch.data.data import GraphData
from graph4nlp.pytorch.modules.prediction.generation.attention import Attention
from graph4nlp.pytorch.modules.prediction.generation.base import RNNDecoderBase
from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab


def extract_mask(mask, token):
    mask_ret = torch.zeros(*(mask.shape)).to(mask.device)
    mask_ret.fill_(0)
    mask_ret[mask == token] = 1
    return mask_ret


class StdRNNDecoder(RNNDecoderBase):
    """
        The standard rnn for sequence decoder.
    Parameters
    ----------
    max_decoder_step: int
        The maximal decoding step.
    input_size: int
        The dimension for standard rnn decoder's input.
    hidden_size: int
        The dimension for standard rnn decoder's hidden representation during calculation.
    word_emb: torch.nn.Embedding
        The target's embedding matrix.
    vocab: Any
        The target's vocabulary
    rnn_type: str, option=["lstm", "gru"], default="lstm"
        The rnn's type. We support ``lstm`` and ``gru`` here.
    use_attention: bool, default=True
        Whether use attention during decoding.
    attention_type: str, option=["uniform", "sep_diff_encoder_type", sep_diff_node_type], default="uniform"
        The attention strategy choice.
        "``uniform``": uniform attention. We will attend on the nodes uniformly.
        "``sep_diff_encoder_type``": separate attention.
            We will attend on graph encoder and rnn encoder's results separately.
        "``sep_diff_node_type``": separate attention.
            We will attend on different node type separately.

    attention_function: str, option=["general", "mlp"], default="mlp"
        Different attention function.
    node_type_num: int, default=None
        When we choose "``sep_diff_node_type``", we must set this parameter.
        This parameter indicate the the amount of node type.
    fuse_strategy: str, option=["average", "concatenate"], default=average
        The strategy to fuse attention results generated by separate attention.
        "``average``": We will take an average on all results.
        "``concatenate``": We will concatenate all results to one.
    use_copy: bool, default=False
        Whether use ``copy`` mechanism. See pointer network. Note that you must use attention first.
    use_coverage: bool, default=False
        Whether use ``coverage`` mechanism. Note that you must use attention first.
    coverage_strategy: str, option=["sum", "max"], default="sum"
        The coverage strategy when calculating the coverage vector.
    tgt_emb_as_output_layer: bool, default=False
        When this option is set ``True``, the output projection layer(It is used to project RNN encoded
        representation to target sequence)'s weight will be shared with the target vocabulary's embedding.
    dropout: float, default=0.3
    """

    def __init__(self, max_decoder_step, input_size, hidden_size,  # decoder config
                 word_emb, vocab: Vocab,  # word embedding & vocabulary TODO: add our vocabulary when building pipeline
                 rnn_type="lstm", graph_pooling_strategy=None,  # RNN config
                 use_attention=True, attention_type="uniform", rnn_emb_input_size=None,  # attention config
                 attention_function="mlp", node_type_num=None, fuse_strategy="average",
                 use_copy=False, use_coverage=False, coverage_strategy="sum",
                 tgt_emb_as_output_layer=False,  # share label projection with word embedding
                 dropout=0.3):
        super(StdRNNDecoder, self).__init__(use_attention=use_attention, use_copy=use_copy, use_coverage=use_coverage,
                                            attention_type=attention_type, fuse_strategy=fuse_strategy)
        self.max_decoder_step = max_decoder_step
        self.word_emb_size = word_emb.embedding_dim
        self.decoder_input_size = input_size
        self.graph_pooling_strategy = graph_pooling_strategy
        self.dropout = nn.Dropout(p=dropout)
        self.tgt_emb = word_emb
        self.tgt_emb_as_output_layer = tgt_emb_as_output_layer
        if self.tgt_emb_as_output_layer:
            self.input_feed_size = self.word_emb_size
        else:
            self.input_feed_size = hidden_size

        self.rnn = self._build_rnn(rnn_type=rnn_type, input_size=self.word_emb_size + self.input_feed_size, hidden_size=hidden_size)
        self.decoder_hidden_size = hidden_size
        self.rnn_type = rnn_type
        self.num_layers = 1
        self.out_logits_size = self.decoder_hidden_size

        # attention builder
        if self.use_attention:
            if self.rnn_type == "lstm":
                query_size = 2 * self.decoder_hidden_size
            elif self.rnn_type == "gru":
                query_size = self.decoder_hidden_size
            else:
                raise NotImplementedError()
            if attention_type == "uniform":
                self.enc_attention = Attention(hidden_size=self.decoder_hidden_size,
                                               query_size=query_size,
                                               memory_size=input_size, has_bias=True,
                                               attention_funtion=attention_function)
                self.out_logits_size += self.decoder_input_size
            elif attention_type == "sep_diff_encoder_type":
                assert isinstance(rnn_emb_input_size, int)
                self.rnn_emb_input_size = rnn_emb_input_size
                self.enc_attention = Attention(hidden_size=self.decoder_hidden_size,
                                               query_size=query_size,
                                               memory_size=input_size, has_bias=True,
                                               attention_funtion=attention_function)
                self.out_logits_size += self.decoder_input_size
                self.rnn_attention = Attention(hidden_size=self.decoder_hidden_size, query_size=query_size,
                                               memory_size=rnn_emb_input_size, has_bias=True,
                                               attention_funtion=attention_function)
                if self.fuse_strategy == "concatenate":
                    self.out_logits_size += self.rnn_emb_input_size
                else:
                    if rnn_emb_input_size != input_size:
                        raise ValueError("input RNN embedding size is not equal to graph embedding size")
            elif attention_type == "sep_diff_node_type":
                assert node_type_num >= 1
                attn_modules = [Attention(hidden_size=self.decoder_hidden_size, query_size=query_size,
                                          memory_size=input_size, has_bias=True,
                                          attention_funtion=attention_function)
                                for _ in range(node_type_num)]
                self.node_type_num = node_type_num
                self.attn_modules = nn.ModuleList(attn_modules)
                if self.fuse_strategy == "concatenate":
                    self.out_logits_size += self.decoder_input_size * self.node_type_num
                elif self.fuse_strategy == "average":
                    self.out_logits_size += self.decoder_input_size
                else:
                    raise NotImplementedError()
            else:
                raise NotImplementedError()

        self.attention_type = attention_type

        if self.rnn_type == "lstm":
            self.encoder_decoder_adapter = nn.ModuleList(
                [nn.Linear(self.decoder_input_size, self.decoder_hidden_size) for _ in range(2)])
        elif self.rnn_type == "gru":
            self.encoder_decoder_adapter = nn.Linear(self.decoder_input_size, self.decoder_hidden_size)
        else:
            raise NotImplementedError()

        # input_feed

        self.memory_to_feed = nn.Sequential(nn.Linear(self.out_logits_size, self.input_feed_size),
                                            nn.ReLU(),
                                            nn.Dropout(dropout))


        # project logits to labels
        # self.tgt_emb_as_output_layer = tgt_emb_as_output_layer
        if self.tgt_emb_as_output_layer:  # use pre_out layer
            self.out_embed_size = self.word_emb_size
            self.pre_out = nn.Linear(self.input_feed_size, self.out_embed_size)
            size_before_output = self.out_embed_size
        else:  # don't use pre_out layer
            size_before_output = self.input_feed_size

        # size_before_output = self.input_feed_size
        self.vocab = vocab
        vocab_size = len(vocab)
        self.vocab_size = vocab_size
        self.out_project = nn.Linear(size_before_output, vocab_size, bias=False)
        if self.tgt_emb_as_output_layer:
            # self.out_project.weight = self.tgt_emb.word_emb_layer.weight
            self.out_project.weight = self.tgt_emb.weight



        # coverage strategy
        if self.use_coverage:
            if not self.use_attention:
                raise ValueError("You should use attention when you use coverage strategy.")

            self.coverage_strategy = coverage_strategy

            def get_coverage_vector(enc_attn_weights):
                if coverage_strategy == 'max':
                    coverage_vector, _ = torch.max(torch.cat(enc_attn_weights), dim=0)
                elif coverage_strategy == 'sum':
                    coverage_vector = torch.sum(torch.cat(enc_attn_weights), dim=0)
                else:
                    raise ValueError('Unrecognized cover_func: ' + self.cover_func)
                return coverage_vector

            self.coverage_function = get_coverage_vector

            self.coverage_weight = torch.Tensor(1, 1, self.decoder_hidden_size)
            self.coverage_weight = nn.Parameter(nn.init.xavier_uniform_(self.coverage_weight))

        # copy: pointer network
        if self.use_copy:
            ptr_size = self.input_feed_size + self.word_emb_size
            if self.rnn_type == "lstm":
                ptr_size += self.decoder_hidden_size * 2
            elif self.rnn_type == "gru":
                ptr_size += self.decoder_hidden_size
            else:
                raise NotImplementedError()
            if self.use_attention:
                if self.attention_type == "uniform":
                    ptr_size += input_size
                elif self.attention_type == "sep_diff_encoder_type":
                    ptr_size += input_size + rnn_emb_input_size
                elif self.attention_type == "sep_diff_node_type":
                    ptr_size += input_size * node_type_num
            self.ptr = nn.Linear(ptr_size, 1)

    def _build_rnn(self, rnn_type, **kwargs):
        """
            The rnn factory.
        Parameters
        ----------
        rnn_type: str, option=["lstm", "gru"], default="lstm"
            The rnn type.
        """
        if rnn_type == 'lstm':
            return nn.LSTM(**kwargs)
        elif rnn_type == "gru":
            return nn.GRU(**kwargs)
        else:
            raise NotImplementedError("RNN type: {} is not supported.".format(rnn_type))

    def _run_forward_pass(self, graph_node_embedding, graph_node_mask=None, rnn_node_embedding=None,
                          graph_level_embedding=None,
                          graph_edge_embedding=None, graph_edge_mask=None, tgt_seq=None, src_seq=None, oov_dict=None,
                          teacher_forcing_rate=1.0):
        """
            The forward function for RNN.
        Parameters
        ----------
        graph_node_embedding: torch.Tensor
            shape=[B, N, D]
        graph_node_mask: torch.Tensor
            shape=[B, N]
            -1 indicating dummy node. 0-``node_type_num`` are valid node type.
        rnn_node_embedding: torch.Tensor
            shape=[B, N, D]
        graph_level_embedding: torch.Tensor
            shape=[B, D]
        graph_edge_embedding: torch.Tensor
            shape=[B, E, D]
            Not implemented yet.
        graph_edge_mask: torch.Tensor
            shape=[B, E]
            Not implemented yet.
        tgt_seq: torch.Tensor
            shape=[B, T]
            The target sequence's index.
        src_seq: torch.Tensor
            shape=[B, S]
            The source sequence's index. It is used for ``use_copy``. Note that it can be encoded by target word
            embedding.
        oov_dict: Vocab
        teacher_forcing_rate: float, default=1.0
            The teacher forcing rate.

        Returns
        -------
        logits: torch.Tensor
            shape=[B, tgt_len, vocab_size]
            The probability for predicted target sequence. It is processed by softmax function.
        enc_attn_weights_average: torch.Tensor
            It is used for calculating coverage loss.
            The averaged attention scores.
        coverage_vectors: torch.Tensor
            It is used for calculating coverage loss.
            The coverage vector.
        """
        target_len = self.max_decoder_step
        if tgt_seq is not None:
            target_len = tgt_seq.shape[1]

        batch_size = graph_node_embedding.shape[0]
        decoder_input = torch.tensor([self.vocab.SOS] * batch_size).to(graph_node_embedding.device)

        decoder_state = self.get_decoder_init_state(rnn_type=self.rnn_type, batch_size=batch_size,
                                                    content=graph_level_embedding)

        input_feed = torch.zeros(batch_size, self.input_feed_size).to(graph_node_embedding.device)


        outputs = []
        enc_attn_weights_average = []
        coverage_vectors = []

        for i in range(target_len):
            decoder_output, input_feed, decoder_state, dec_attn_scores, coverage_vec = \
                self.decode_step(decoder_input=decoder_input, input_feed=input_feed, rnn_state=decoder_state, dec_input_mask=graph_node_mask,
                                 encoder_out=graph_node_embedding, rnn_emb=rnn_node_embedding,
                                 enc_attn_weights_average=enc_attn_weights_average,
                                 src_seq=src_seq, oov_dict=oov_dict)
            if self.use_coverage:
                enc_attn_weights_average.append(dec_attn_scores.unsqueeze(0))
                coverage_vectors.append(coverage_vec)

            outputs.append(decoder_output.unsqueeze(1))

            # teacher_forcing
            if tgt_seq is not None and random.random() < teacher_forcing_rate:
                decoder_input = tgt_seq[:, i]
            else:
                # sampling
                # TODO: now argmax sampling
                decoder_input = decoder_output.squeeze(1).argmax(dim=-1)
            decoder_input = self._filter_oov(decoder_input)
        ret = torch.cat(outputs, dim=1)

        return ret, enc_attn_weights_average, coverage_vectors

    def decode_step(self, decoder_input, input_feed, rnn_state, encoder_out, dec_input_mask, rnn_emb=None,
                    enc_attn_weights_average=None,
                    src_seq=None, oov_dict=None):
        batch_size = decoder_input.shape[0]
        dec_emb = self.tgt_emb(decoder_input)
        dec_emb = torch.cat((dec_emb, input_feed), dim=1)
        dec_emb = self.dropout(dec_emb)
        if self.use_coverage and enc_attn_weights_average:
            coverage_vec = self.coverage_function(enc_attn_weights_average)
        else:
            coverage_vec = None

        dec_out, rnn_state = self.rnn(dec_emb.unsqueeze(0), rnn_state)
        dec_out = dec_out.squeeze(0)

        if self.rnn_type == "lstm":
            rnn_state = tuple([self.dropout(x) for x in rnn_state])
            hidden = torch.cat(rnn_state, -1).squeeze(0)
        elif self.rnn_type == "gru":
            rnn_state = self.dropout(rnn_state)
            hidden = rnn_state.squeeze(0)
        else:
            raise NotImplementedError()
        attn_collect = []
        score_collect = []

        if self.use_attention:
            if self.use_coverage and coverage_vec is not None:
                coverage_repr = coverage_vec
            else:
                coverage_repr = None
            if coverage_repr is not None:
                coverage_repr = coverage_repr.unsqueeze(-1) * self.coverage_weight
            if self.attention_type == "uniform" or self.attention_type == "sep_diff_encoder_type":
                enc_mask = extract_mask(dec_input_mask, token=-1)
                enc_mask = 1 - enc_mask

                attn_res, scores = self.enc_attention(query=hidden, memory=encoder_out, memory_mask=enc_mask,
                                                      coverage=coverage_repr)
                attn_collect.append(attn_res)
                score_collect.append(scores)
                if self.attention_type == "sep_diff_encoder_type":
                    rnn_attn_res, rnn_scores = self.rnn_attention(query=hidden, memory=rnn_emb, memory_mask=enc_mask, coverage=coverage_repr)
                    score_collect.append(rnn_scores)
                    attn_collect.append(rnn_attn_res)
            elif self.attention_type == "sep_diff_node_type":
                for i in range(self.node_type_num):
                    node_mask = extract_mask(dec_input_mask, token=i)
                    attn, scores = self.attn_modules[i](query=hidden, memory=encoder_out, memory_mask=node_mask,
                                                        coverage=coverage_repr)
                    attn_collect.append(attn)
                    score_collect.append(scores)

        if self.use_attention:
            if self.attention_type == "uniform":
                assert len(attn_collect) == 1
                assert len(score_collect) == 1
                attn_total = attn_collect[0]
            elif self.attention_type == "sep_diff_encoder_type" or self.attention_type == "sep_diff_node_type":
                if self.fuse_strategy == "average":
                    attn_total = reduce(lambda x, y: x + y, attn_collect) / len(attn_collect)
                elif self.fuse_strategy == "concatenate":
                    attn_total = torch.cat(attn_collect, dim=-1)
                else:
                    raise NotImplementedError()
            else:
                raise NotImplementedError()

            decoder_output = torch.cat((dec_out, attn_total), dim=-1)
            dec_attn_scores = reduce(lambda x, y: x + y, score_collect) / len(score_collect)
        else:
            decoder_output = dec_out

        out = self.memory_to_feed(decoder_output)

        # project
        if self.tgt_emb_as_output_layer:
            out_embed = torch.tanh(self.pre_out(out))
            # out_embed = torch.tanh(out)
            out_embed = self.dropout(out_embed)
        else:
            out_embed = out

        decoder_output = self.out_project(out_embed)  # [B, S, Vocab]

        if self.use_copy:
            assert src_seq is not None
            assert oov_dict is not None
            #output = torch.zeros(batch_size, oov_dict.get_vocab_size()).to(decoder_input.device)
            attn_ptr = torch.cat(attn_collect, dim=-1)
            pgen_collect = [dec_emb, hidden, attn_ptr]

            prob_ptr = torch.sigmoid(self.ptr(torch.cat(pgen_collect, -1)))
            prob_gen = 1 - prob_ptr
            gen_output = torch.softmax(decoder_output, dim=-1)

            ret = prob_gen * gen_output
            need_pad_length = oov_dict.get_vocab_size() - self.vocab.get_vocab_size()
            output = torch.cat((ret, ret.new_zeros((batch_size, need_pad_length))), dim=1)
            # output[:, :self.vocab.get_vocab_size()] = ret

            ptr_output = dec_attn_scores

            output.scatter_add_(1, src_seq, prob_ptr * ptr_output)
            decoder_output = output
        else:
            decoder_output = torch.softmax(decoder_output, dim=-1)

        return decoder_output, out, rnn_state, dec_attn_scores, coverage_vec

    def get_decoder_init_state(self, rnn_type, batch_size, content=None):
        if rnn_type == "lstm":
            if content is not None:
                assert len(content.shape) == 2
                assert content.shape[0] == batch_size
                ret = tuple([self.encoder_decoder_adapter[i](content).view(1, batch_size,
                                                                           self.decoder_hidden_size).expand(
                    self.num_layers, -1, -1) for i in range(2)])
            else:
                weight = next(self.parameters()).data
                ret = (weight.new(self.num_layers, batch_size, self.decoder_hidden_size).zero_(),
                       weight.new(self.num_layers, batch_size, self.decoder_hidden_size).zero_())
        elif rnn_type == "gru":
            if content is not None:
                ret = self.encoder_decoder_adapter(content).view(1, batch_size,
                                                                 self.decoder_hidden_size).expand(
                    self.num_layers, -1, -1)
            else:
                weight = next(self.parameters()).data
                ret = weight.new(self.num_layers, batch_size, self.decoder_hidden_size).zero_()
        else:
            raise NotImplementedError()
        return ret

    def _filter_oov(self, tokens):
        ret = tokens.clone()
        ret[tokens >= self.vocab_size] = self.vocab.UNK
        return ret

    def forward(self, batch_graph, tgt_seq=None, oov_dict=None, teacher_forcing_rate=1.0):
        """
            The forward function of ``StdRNNDecoder``
        Parameters
        ----------
        batch_graph: GraphData
            The graph input
        tgt_seq: torch.Tensor
            shape=[B, T]
            The target sequence's index.
        oov_dict: VocabModel, default=None
            The vocabulary for copy mechanism.
        teacher_forcing_rate: float, default=1.0
            The teacher forcing rate.

        Returns
        -------
        logits: torch.Tensor
            shape=[B, tgt_len, vocab_size]
            The probability for predicted target sequence. It is processed by softmax function.
        enc_attn_weights_average: torch.Tensor
            It is used for calculating coverage loss.
            The averaged attention scores.
        coverage_vectors: torch.Tensor
            It is used for calculating coverage loss.
            The coverage vector.
        """
        params = self.extract_params(batch_graph)
        params['tgt_seq'] = tgt_seq
        params['teacher_forcing_rate'] = teacher_forcing_rate
        params["oov_dict"] = oov_dict
        return self._run_forward_pass(**params)

    def extract_params(self, batch_graph):
        """
            Extract parameters from ``batch_graph`` for _run_forward_pass() function.
        Parameters
        ----------
        batch_graph: GraphData

        Returns
        -------
        params: dict
        """
        batch_data_dict = batch_graph.batch_node_features
        graph_node_emb = batch_data_dict["node_emb"]

        rnn_node_emb = batch_data_dict["rnn_emb"]

        if len(batch_data_dict["token_id"].shape) == 3:
            graph_node_mask = (torch.sum(batch_data_dict["token_id"], dim=-1) != 0).squeeze(-1).float() - 1
        else:
            graph_node_mask = (batch_data_dict["token_id"] != 0).squeeze(-1).float() - 1

        if self.use_copy:
            src_seq_ret = batch_graph.batch_node_features["token_id_oov"]
        else:
            src_seq_ret = None

        graph_level_emb = self.graph_pooling(graph_node_emb)

        return {
            "graph_node_embedding": graph_node_emb,
            "graph_node_mask": graph_node_mask,
            "rnn_node_embedding": rnn_node_emb,
            "graph_level_embedding": graph_level_emb,
            "graph_edge_embedding": None,
            "graph_edge_mask": None,
            "src_seq": src_seq_ret.long() if self.use_copy else None
        }

    def graph_pooling(self, graph_node):
        if self.graph_pooling_strategy is None:
            pooled_vec = None
        elif self.graph_pooling_strategy == "mean":
            pooled_vec = torch.mean(graph_node, dim=1)
        elif self.graph_pooling_strategy == "max":
            pooled_vec, _ = torch.max(graph_node, dim=1)
        elif self.graph_pooling_strategy == "min":
            pooled_vec, _ = torch.mean(graph_node, dim=1)
        else:
            raise NotImplementedError()
        return pooled_vec



from .base import GeneralLossBase
import torch


class CoverageLoss(GeneralLossBase):
    """
        The loss function for coverage mechanism.
    Parameters
    ----------
    cover_loss: float
        The weight for coverage loss.
    """
    def __init__(self, cover_loss):
        super(CoverageLoss, self).__init__()
        self.cover_loss = cover_loss

    def forward(self, enc_attn_weights, coverage_vectors):
        """
            The calculation function.
        Parameters
        ----------
        enc_attn_weights: list[torch.Tensor]
            The list containing all decoding steps' attention weights.
            The length should be the decoding step.
            Each element should be the tensor.
        coverage_vectors: list[torch.Tensor]
            The list containing all coverage vectors in decoding module.
        Returns
        -------
        coverage_loss: torch.Tensor
            The loss.
        """
        target_length = len(enc_attn_weights)
        loss = 0
        for i in range(target_length):
            if coverage_vectors[i] is not None:
                coverage_loss = torch.sum(
                    torch.min(coverage_vectors[i], enc_attn_weights[i])) / coverage_vectors[-1].shape[0] * self.cover_loss
                loss += coverage_loss
        return loss / target_length


from copy import copy
from scipy import sparse
import torch

from ...data.data import GraphData

CORENLP_TIMEOUT_SIGNATURE = 'CoreNLP request timed out. Your document may be too long.'

def convert_adj_to_graph(graph, adj, reverse_adj, mask_off_val):

    slides = (adj != mask_off_val).nonzero(as_tuple=False)
    from copy import deepcopy
    batch_nodes_tensor = torch.Tensor([0] + graph._batch_num_nodes).to(slides.device)
    batch_prefix = batch_nodes_tensor.view(-1, 1).expand(-1, batch_nodes_tensor.shape[0])

    batch_prefix = batch_prefix.triu().long().sum(0)


    # for i in range(1, len(graph._batch_num_nodes)):
    #     cnt = batch_nodes[-1] + graph._batch_num_nodes[i-1]
    #     batch_nodes.append(cnt)

    # pre_num = [batch_nodes[i] for i in slides[:, 0]]


    src = slides[:, 1] + batch_prefix.index_select(dim=0, index=slides[:, 0])
    tgt = slides[:, 2] + batch_prefix.index_select(dim=0, index=slides[:, 0])

    graph_data = graph
    graph_data.remove_all_edges() # remove all existing edges
    graph_data.add_edges(src.detach().cpu().numpy().tolist(), tgt.detach().cpu().numpy().tolist())

    value = adj[slides[:, 0], slides[:, 1], slides[:, 2]]
    reverse_value = reverse_adj[slides[:, 0], slides[:, 1], slides[:, 2]]
    graph_data.edge_features['edge_weight'] = value
    graph_data.edge_features['reverse_edge_weight'] = reverse_value

    return graph_data





# def convert_adj_to_graph(graph, adj, reverse_adj, mask_off_val):
#     """Convert adjacency matrix to GraphData
#     """
#     # slides = (adj != mask_off_val).nonzero()
#     # from copy import deepcopy
#     # batch_nodes = [0]


#     # for i in range(1, len(graph._batch_num_nodes)):
#     #     cnt = batch_nodes[-1] + graph._batch_num_nodes[i-1]
#     #     batch_nodes.append(cnt)

#     # # pre_num = [batch_nodes[i] for i in slides[:, 0]]


#     # src = slides[:, 1] + slides[:, 0] * 100
#     # tgt = slides[:, 2] + slides[:, 0]* 100

#     # graph_data = graph
#     # graph_data.add_edges(src.detach().cpu().numpy().tolist(), tgt.detach().cpu().numpy().tolist())

#     # value = adj[slides[:, 0], slides[:, 1], slides[:, 2]]
#     # reverse_value = reverse_adj[slides[:, 0], slides[:, 1], slides[:, 2]]
#     # #graph_data.edge_features['edge_weight'] = value
#     # #graph_data.edge_features['reverse_edge_weight'] = reverse_value

#     # return graph_data


#     slides = (adj != mask_off_val).nonzero()
#     src = slides[:, 0]
#     tgt = slides[:, 1]
#     graph.add_edges(src.detach().cpu().numpy().tolist(), tgt.detach().cpu().numpy().tolist())
#     # value = adj[slides[:, 0], slides[:, 1]]
#     # reverse_value = reverse_adj[slides[:, 0], slides[:, 1]]
#     # graph.edge_features['edge_weight'] = value
#     # graph.edge_features['reverse_edge_weight'] = reverse_value
#     return graph




#     binarized_adj = sparse.coo_matrix(adj.detach().cpu().numpy() != mask_off_val)
#     graph_data = GraphData()
#     graph_data.from_scipy_sparse_matrix(binarized_adj)
#     edge_weight = adj[adj != mask_off_val]
#     reverse_edge_weight = reverse_adj[reverse_adj != mask_off_val]

#     graph_data.edge_features['edge_weight'] = edge_weight
#     graph_data.edge_features['reverse_edge_weight'] = reverse_edge_weight

#     return graph_data



from torch import nn
import torch
from ..base import KGCompletionLayerBase


class DistMultLayer(KGCompletionLayerBase):
    r"""Specific class for knowledge graph completion task.

    DistMult from paper `Embedding entities and relations for learning and
    inference in knowledge bases <https://arxiv.org/pdf/1412.6575.pdf>`__.

    .. math::
        f(s, r, o) & = e_s^T R_r e_o

    Parameters
    ----------
    input_dropout: float
        Dropout for node_emb and rel_emb. Default: 0.0

    loss_name: str
        The loss type selected fot the KG completion task.

    """

    def __init__(self,
                 input_dropout=0.0,
                 loss_name='BCELoss'):
        super(DistMultLayer, self).__init__()
        self.inp_drop = nn.Dropout(input_dropout)
        self.loss_name = loss_name

    def forward(self,
                e1_emb,
                rel_emb,
                all_node_emb,
                multi_label=None):
        r"""

        Parameters
        ----------

        e1_emb : tensor [B, H]
            The selected entity_1 embeddings of a batch.
            B: batch size
            H: length of the node embeddings (entity embeddings)

        rel_emb : tensor [B, H]
            The selected relation embeddings of a batch.
            B: batch size
            H: length of the edge embeddings (relation embeddings)

        all_node_emb :  torch.nn.modules.sparse.Embedding [N, H]
            All node embeddings.
            N: number of nodes in the whole KG graph
            H: length of the node embeddings (entity embeddings)

        multi_label: tensor [B, N]
            multi_label is a binary matrix. Each element can be equal to 1 for true label
            and 0 for false label (or 1 for true label, -1 for false label).
            multi_label[i] represents a multi-label of a given head-rel pair.
            B is the batch size.
            N: number of nodes in the whole KG graph.

        Returns
        -------
        pred: tensor [B, N].
            The score logits for all nodes preidcted.

        pred_pos: tensor [B_p]
            The predition scores of positive examples.

        pred_neg: tensor [B_n]
            The predition scores of negative examples.
            B_p + B_n == B * N.

        """
        # dropout
        e1_emb = self.inp_drop(e1_emb)
        rel_emb = self.inp_drop(rel_emb)

        logits = torch.mm(e1_emb * rel_emb,
                          all_node_emb.weight.transpose(1, 0))


        if self.loss_name in ['SoftMarginLoss']:
            # target labels are numbers selecting from -1 and 1.
            pred = torch.tanh(logits)
        else:
            # target labels are numbers selecting from 0 and 1.
            pred = torch.sigmoid(logits)

        if multi_label is not None:
            idxs_pos = torch.nonzero(multi_label == 1.)
            pred_pos = pred[idxs_pos[:, 0], idxs_pos[:, 1]]

            idxs_neg = torch.nonzero(multi_label == 0.)
            pred_neg = pred[idxs_neg[:, 0], idxs_neg[:, 1]]
            return pred, pred_pos, pred_neg
        else:
            return pred


from torch import nn
import torch
from ..base import KGCompletionLayerBase


class ComplExLayer(KGCompletionLayerBase):
    r"""Specific class for knowledge graph completion task.

        ComplEx from paper `Complex Embeddings for Simple Link Prediction
        <http://proceedings.mlr.press/v48/trouillon16.pdf>`__.

        Parameters
        ----------
        input_dropout: float
           Dropout for node_emb and rel_emb. Default: 0.0

        loss_name: str
           The loss type selected fot the KG completion task.

    """
    def __init__(self,
                 input_dropout=0.0,
                 loss_name='BCELoss'):
        super(ComplExLayer, self).__init__()
        self.inp_drop = nn.Dropout(input_dropout)
        self.loss_name = loss_name

    def forward(self,
                e1_embedded_real,
                e1_embedded_img,
                rel_embedded_real,
                rel_embedded_img,
                all_node_emb_real,
                all_node_emb_img,
                multi_label=None):
        r"""

        Parameters
        ----------

        input graph : GraphData
            The tensors stored in the node feature field named "node_emb" and
            "rel_emb" in the input_graph are used for knowledge graph completion.

        e1_embedded_real : tensor [B, H]
            The selected entity_1 real embeddings of a batch.
            B: batch size
            H: length of the node embeddings (entity embeddings)

        rel_embedded_real : tensor [B, H]
            The selected relation real embeddings of a batch.
            B: batch size
            H: length of the edge embeddings (relation embeddings)

        e1_embedded_img : tensor [B, H]
            The selected entity_1 img embeddings of a batch.
            B: batch size
            H: length of the node embeddings (entity embeddings)

        rel_embedded_img : tensor [B, H]
            The selected relation img embeddings of a batch.
            B: batch size
            H: length of the edge embeddings (relation embeddings)

        all_node_emb_real :  torch.nn.modules.sparse.Embedding [N, H]
            All node real embeddings.
            N: number of nodes in the whole KG graph
            H: length of the node real embeddings (entity embeddings)

        all_node_emb_img :  torch.nn.modules.sparse.Embedding [N, H]
            All node img embeddings.
            N: number of nodes in the whole KG graph
            H: length of the node img embeddings (entity embeddings)

        multi_label: tensor [B, N]
            multi_label is a binary matrix. Each element can be equal to 1 for true label
            and 0 for false label (or 1 for true label, -1 for false label).
            multi_label[i] represents a multi-label of a given head-rel pair.
            B is the batch size.
            N: number of nodes in the whole KG graph.

        Returns
        -------
        pred: tensor [B, N].
            The score logits for all nodes preidcted.

        pred_pos: tensor [B_p]
            The predition scores of positive examples.

        pred_neg: tensor [B_n]
            The predition scores of negative examples.
            B_p + B_n == B * N.

        """
        e1_embedded_real = self.inp_drop(e1_embedded_real)
        rel_embedded_real = self.inp_drop(rel_embedded_real)
        e1_embedded_img = self.inp_drop(e1_embedded_img)
        rel_embedded_img = self.inp_drop(rel_embedded_img)

        # complex space bilinear product (equivalent to HolE)
        realrealreal = torch.mm(e1_embedded_real * rel_embedded_real, all_node_emb_real.weight.transpose(1, 0))
        realimgimg = torch.mm(e1_embedded_real * rel_embedded_img, all_node_emb_img.weight.transpose(1, 0))
        imgrealimg = torch.mm(e1_embedded_img * rel_embedded_real, all_node_emb_img.weight.transpose(1, 0))
        imgimgreal = torch.mm(e1_embedded_img * rel_embedded_img, all_node_emb_real.weight.transpose(1, 0))
        pred = realrealreal + realimgimg + imgrealimg - imgimgreal

        if self.loss_name in ["SoftMarginLoss"]:
            pred = torch.tanh(pred)
        else:
            pred = torch.sigmoid(pred)

        if multi_label is not None:
            idxs_pos = torch.nonzero(multi_label == 1.)
            pred_pos = pred[idxs_pos[:, 0], idxs_pos[:, 1]]

            idxs_neg = torch.nonzero(multi_label == 0.)
            pred_neg = pred[idxs_neg[:, 0], idxs_neg[:, 1]]

            return pred, pred_pos, pred_neg
        else:
            return pred


from .BiLSTMFeedForwardNN import BiLSTMFeedForwardNN
from .BiLSTMFeedForwardNNLayer import BiLSTMFeedForwardNNLayer
from .FeedForwardNN import FeedForwardNN
from .FeedForwardNNLayer import FeedForwardNNLayer

__all__ = ['BiLSTMFeedForwardNN', 'BiLSTMFeedForwardNNLayer', 
            'FeedForwardNN', 'FeedForwardNNLayer']



from nltk.tokenize import word_tokenize
import torch
from torch import nn

from ...data.data import GraphData
from .base import DynamicGraphConstructionBase
from .dependency_graph_construction import DependencyBasedGraphConstruction
from .constituency_graph_construction import ConstituencyBasedGraphConstruction
from .ie_graph_construction import IEBasedGraphConstruction
from ..utils.generic_utils import normalize_adj, to_cuda
# from ..utils.constants import VERY_SMALL_NUMBER
from .utils import convert_adj_to_graph
from ...data.data import to_batch
from ..utils.vocab_utils import Vocab


class NodeEmbeddingBasedRefinedGraphConstruction(DynamicGraphConstructionBase):
    """Class for node embedding based refined dynamic graph construction.

    Parameters
    ----------
    word_vocab : Vocab
        The word vocabulary.
    embedding_styles : dict
        - ``single_token_item`` : specify whether the item (i.e., node or edge) contains single token or multiple tokens.
        - ``emb_strategy`` : specify the embedding construction strategy.
        - ``num_rnn_layers``: specify the number of RNN layers.
        - ``bert_model_name``: specify the BERT model name.
        - ``bert_lower_case``: specify whether to lower case the input text for BERT embeddings.
    alpha_fusion : float
        Specify the fusion value for combining initial and learned adjacency matrices.
    """
    def __init__(self,
                word_vocab,
                embedding_styles,
                alpha_fusion,
                **kwargs):
        super(NodeEmbeddingBasedRefinedGraphConstruction, self).__init__(
                                                            word_vocab,
                                                            embedding_styles,
                                                            **kwargs)
        assert 0 <= alpha_fusion <= 1, 'alpha_fusion should be a `float` number between 0 and 1'
        self.alpha_fusion = alpha_fusion

    def forward(self, batch_graphdata):
        """Compute graph topology and initial node embeddings.

        Parameters
        ----------
        batch_graphdata : GraphData
            The input graph data.

        Returns
        -------
        GraphData
            The constructed batched graph.
        """
        init_norm_adj = self._get_normalized_init_adj(batch_graphdata)
        batch_graphdata = self.embedding(batch_graphdata)
        batch_graphdata = self.topology(batch_graphdata, init_norm_adj)

        return batch_graphdata


    def topology(self, graph, init_norm_adj):
        """Compute graph topology.

        Parameters
        ----------
        graph : GraphData
            The input graph data.
        init_norm_adj : torch.sparse.FloatTensor
            The initial init_norm_adj adjacency matrix.

        Returns
        -------
        GraphData
            The constructed graph.
        """
        node_emb = graph.batch_node_features["node_feat"]
        node_mask = (graph.batch_node_features["token_id"] != Vocab.PAD)

        raw_adj = self.compute_similarity_metric(node_emb, node_mask)
        raw_adj = self.sparsify_graph(raw_adj)
        graph_reg = self.compute_graph_regularization(raw_adj, node_emb)

        if self.sim_metric_type in ('rbf_kernel', 'weighted_cosine'):
            assert raw_adj.min().item() >= 0, 'adjacency matrix must be non-negative!'
            adj = raw_adj / torch.clamp(torch.sum(raw_adj, dim=-1, keepdim=True), min=torch.finfo(torch.float32).eps)
            reverse_adj = raw_adj / torch.clamp(torch.sum(raw_adj, dim=-2, keepdim=True), min=torch.finfo(torch.float32).eps)
        elif self.sim_metric_type == 'cosine':
            raw_adj = (raw_adj > 0).float()
            adj = normalize_adj(raw_adj)
            reverse_adj = adj
        else:
            adj = torch.softmax(raw_adj, dim=-1)
            reverse_adj = torch.softmax(raw_adj, dim=-2)

        if self.alpha_fusion is not None:
            adj = torch.sparse.FloatTensor.add((1 - self.alpha_fusion) * adj, self.alpha_fusion * init_norm_adj)
            reverse_adj = torch.sparse.FloatTensor.add((1 - self.alpha_fusion) * reverse_adj, self.alpha_fusion * init_norm_adj)

        graph = convert_adj_to_graph(graph, adj, reverse_adj, 0)
        graph.graph_attributes['graph_reg'] = graph_reg

        return graph

    @classmethod
    def init_topology(cls,
                    raw_text_data,
                    lower_case=True,
                    tokenizer=word_tokenize,
                    nlp_processor=None,
                    processor_args=None,
                    merge_strategy=None,
                    edge_strategy=None,
                    verbase=False,
                    dynamic_init_topology_builder=None,
                    dynamic_init_topology_aux_args=None):
        """Convert raw text data to the initial graph.

        Parameters
        ----------
        raw_text_data : str or list/tuple of str
            The raw text data. When a list/tuple of tokens is provided, no
            tokenization will be conducted and each token is a node
            (used for line graph builder); otherwise, tokenization will
            be conducted on the input string to get a list of tokens.
        lower_case : boolean
            Specify whether to lower case the input text, default: ``True``.
        tokenizer : callable, optional
            The tokenization function, default: ``nltk.tokenize.word_tokenize``.
        nlp_processor: StanfordCoreNLP, optional
            The NLP processor, default: ``None``.
        processor_args: dict, optional
            The NLP processor arguments, default: ``None``.
        merge_strategy: str
            Strategy to merge sub-graphs into one graph, depends on specific ``dynamic_init_topology_builder``, default: ``None``.
        edge_strategy: str
            Strategy to process edge, depends on specific ``dynamic_init_topology_builder``, default: ``None``.
        verbase: boolean
            verbase flag, default: ``False``.
        dynamic_init_topology_builder : class, optional
            The initial graph topology builder, default: ``None``.
        dynamic_init_topology_aux_args : dict, optional
            The auxiliary args for dynamic_init_topology_builder.topology, default: ``None``.

        Returns
        -------
        GraphData
            The constructed graph.
        """
        if dynamic_init_topology_builder is None: # line graph
            if isinstance(raw_text_data, str):
                token_list = tokenizer(raw_text_data.strip())
            elif isinstance(raw_text_data, (list, tuple)):
                token_list = raw_text_data
            else:
                raise RuntimeError('raw_text_data must be str or list/tuple of str')

            graph = GraphData()
            graph.add_nodes(len(token_list))

            for idx in range(len(token_list) - 1):
                graph.add_edge(idx, idx + 1)
                graph.add_edge(idx + 1, idx)
                graph.node_attributes[idx]['token'] = token_list[idx].lower() if lower_case else token_list[idx]

            graph.node_attributes[idx + 1]['token'] = token_list[-1].lower() if lower_case else token_list[-1]
        elif dynamic_init_topology_builder in (IEBasedGraphConstruction, DependencyBasedGraphConstruction, ConstituencyBasedGraphConstruction):
            graph = dynamic_init_topology_builder.topology(
                                raw_text_data=raw_text_data,
                                nlp_processor=nlp_processor,
                                processor_args=processor_args,
                                merge_strategy=merge_strategy,
                                edge_strategy=edge_strategy,
                                verbase=verbase)
        else:
            graph = dynamic_init_topology_builder.topology(
                                raw_text_data,
                                dynamic_init_topology_aux_args)

        return graph



import copy


class Node():
    """
    Class for self-defined node data structure
    
    ...
    
    Attributes
    ----------
    word_ : (str)
        The text contained in this node.
    
    type_ : (int)
        Node type in graph data structure, 0 for word nodes, 1 for constituency nodes, 2~n can be defined by other node types.
    
    id_ : (int)
        Unique identifier for nodes.

    sentence_ : (int)
        To illustrate which sentence the node is related when multiple sentences included in a graph.
    """
    def __init__(self, word_, type_, id_, sentence_):
        # word: this node's text
        self.word = word_

        # type: 0 for word nodes, 1 for constituency nodes
        self.type = type_

        # id: unique identifier for every node
        self.id = id_

        self.head = False

        self.tail = False

        self.sentence = sentence_

    def __str__(self):
        return self.word + "_type_" + str(self.type) + "_id_" + str(
            self.id) + "_head_" + str(self.head) + "_tail_" + str(
                self.tail) + "_sentence_" + str(self.sentence)


class UtilityFunctionsForGraph():
    """
    A class for utility functions of graph operations.
    """
    @staticmethod
    def get_head_node(g, sentence_id):
        """Return the head node in a syntactic graph, to be used when we need to connect multiple graphs into one merged graph."""
        for n in g.nodes():
            if (n.head == True) and (n.sentence == sentence_id):
                return n

    @staticmethod
    def get_tail_node(g, sentence_id):
        """Return the tail node in a syntactic graph, to be used when we need to connect multiple graphs into one merged graph."""
        for n in g.nodes():
            if (n.tail == True) and (n.sentence == sentence_id):
                return n

    @staticmethod
    def cut_root_node(g):
        """Remove the \"Root\" node in a graph."""
        pass

    @staticmethod
    def cut_pos_node(g):
        """Remove some part-of-speech nodes in a graph, e.g., \"NN\", \"JJ\"."""
        node_arr = list(g.nodes())
        del_arr = []
        for n in node_arr:
            edge_arr = list(g.edges())
            cnt_in = 0
            cnt_out = 0
            for e in edge_arr:
                if n.id == e[0].id:
                    cnt_out += 1
                    out_ = e[1]
                if n.id == e[1].id:
                    cnt_in += 1
                    in_ = e[0]
            if cnt_in == 1 and cnt_out == 1 and out_.type == 0:
                del_arr.append((n, in_, out_))
        for d in del_arr:
            g.remove_node(d[0])
            g.add_edge(d[1], d[2])
        return g

    @staticmethod
    def cut_line_node(g):
        """Remove intermediate nodes organized like a line."""
        node_arr = list(g.nodes())

        for n in node_arr:
            edge_arr = list(g.edges())
            cnt_in = 0
            cnt_out = 0
            for e in edge_arr:
                if n.id == e[0].id:
                    cnt_out += 1
                    out_ = e[1]
                if n.id == e[1].id:
                    cnt_in += 1
                    in_ = e[0]
            if cnt_in == 1 and cnt_out == 1:
                g.remove_node(n)
                g.add_edge(in_, out_)
        return g

    @staticmethod
    def get_seq_nodes(g):
        """Return word nodes in a syntactic graph."""
        res = []
        node_arr = list(g.nodes())
        for n in node_arr:
            if n.type == 0:
                res.append(copy.deepcopy(n))
        return sorted(res, key=lambda x: x.id)

    @staticmethod
    def get_non_seq_nodes(g):
        """Return non-word nodes (like constituency tag nodes) in a syntactic graph."""
        res = []
        node_arr = list(g.nodes())
        for n in node_arr:
            if n.type != 0:
                res.append(copy.deepcopy(n))
        return sorted(res, key=lambda x: x.id)

    @staticmethod
    def get_all_text(g):
        """Return all nodes text strings in a string list."""
        seq_arr = UtilityFunctionsForGraph.get_seq_nodes(g)
        nonseq_arr = UtilityFunctionsForGraph.get_non_seq_nodes(g)
        seq = [x.word for x in seq_arr]
        nonseq = [x.word for x in nonseq_arr]
        return seq + nonseq

    @staticmethod
    def get_all_id(g):
        """Return all nodes id."""
        seq_arr = UtilityFunctionsForGraph.get_seq_nodes(g)
        nonseq_arr = UtilityFunctionsForGraph.get_non_seq_nodes(g)
        seq = [x.id for x in seq_arr]
        nonseq = [x.id for x in nonseq_arr]
        return seq + nonseq

    @staticmethod
    def get_id2word(g):
        """Return a dict for mapping from nodes id to word text."""
        res = {}
        seq_arr = UtilityFunctionsForGraph.get_seq_nodes(g)
        nonseq_arr = UtilityFunctionsForGraph.get_non_seq_nodes(g)
        for x in seq_arr:
            res[x.id] = x.word
        for x in nonseq_arr:
            res[x.id] = x.word
        return res

    @staticmethod
    def print_edges(g):
        """Display all edges text and id in a graph."""
        edge_arr = list(g.edges())
        for e in edge_arr:
            print(e[0].word, e[1].word), (e[0].id, e[1].id)

    @staticmethod
    def print_nodes(g):
        """Display all nodes in a graph."""
        nodes_arr = list(g.nodes())
        for n in nodes_arr:
            print(n)



import copy
import torch
import numpy as np

from .padding_utils import pad_2d_vals_no_size
from ...data.data import GraphData
from .vocab_utils import VocabModel, Vocab


def prepare_ext_vocab(batch_graph, vocab, gt_str=None, device=None):
    """
        The function for copy mechanism.
        It will find the oov dict in ``gt_str`` and build ``oov_dict``.
        And the tgt_seq will be built by ``oov_dict``.
        Note that, the word_index in ``oov_dict`` is saved in ``token_id_oov`` of GraphData.
    Parameters
    ----------
    batch_graph: GraphData
        The graph.
    vocab: VocabModel
        The VocabModel.
    gt_str: list[str], default=None
        The list contains the raw target (i.e. the ground-truth for most task) sequence.
        Note that the raw sequence should not contains <UNK>.
        It will be processed by ``oov_dict`` and generate the ``tgt_seq``.
    device: torch.device, default=None
        The device.
    Returns
    -------
    oov_dict: Vocab
        The vocab contains the out-of-vocabulary words in batch_graph.
    tgt_seq: torch.Tensor
        If the ``gt_str`` is not ``None``, it will be calculated.
        We process the ``gt_str`` with ``oov_dict``.
    """
    oov_dict = copy.deepcopy(vocab.in_word_vocab)
    token_matrix = batch_graph.node_features["token_id"].squeeze(1)
    unk_index = (token_matrix == oov_dict.UNK).nonzero(as_tuple=False).squeeze(1).detach().cpu().numpy()
    # unk_token = batch_graph.node_attributes[unk_index.squeeze(1).detach().cpu().numpy().tolist()[:]]["token"]
    unk_token = [batch_graph.node_attributes[index]["token"] for index in unk_index]
    oov_dict._add_words(unk_token)
    token_matrix_oov = token_matrix.clone()
    for idx in unk_index:
        unk_token = batch_graph.node_attributes[idx]["token"]
        oov_dict._add_words(unk_token)
        token_matrix_oov[idx] = oov_dict.getIndex(unk_token)
    batch_graph.node_features["token_id_oov"] = token_matrix_oov


    if gt_str is not None:
        oov_tgt_collect = []
        for s in gt_str:
            oov_tgt = oov_dict.to_index_sequence(s)
            oov_tgt.append(oov_dict.EOS)
            oov_tgt = np.array(oov_tgt)
            oov_tgt_collect.append(oov_tgt)

        output_pad = pad_2d_vals_no_size(oov_tgt_collect)

        tgt_seq = torch.from_numpy(output_pad).long().to(device)
        return oov_dict, tgt_seq
    else:
        return oov_dict


import torch
import dgl
from dgl.nn.pytorch.glob import AvgPooling as DGLAvgPooling

from ..base import PoolingBase
from .....data.data import from_batch


class AvgPooling(PoolingBase):
    r"""Apply average pooling over the nodes in the graph.

    .. math::
        r^{(i)} = \frac{1}{N_i}\sum_{k=1}^{N_i} x^{(i)}_k
    """
    def __init__(self):
        super(AvgPooling, self).__init__()
        # self.model = DGLAvgPooling()

    def forward(self, graph, feat):
        r"""Compute average pooling.

        Parameters
        ----------
        graph : GraphData
            The graph data.
        feat : str
            The feature field name.

        Returns
        -------
        torch.Tensor
            The output feature.
        """
        graph_list = from_batch(graph)
        output_feat = []
        for g in graph_list:
            output_feat.append(g.node_features[feat].mean(dim=0))

        output_feat = torch.stack(output_feat, 0)

        return output_feat



import torch
import torch.nn as nn
import torch.nn.functional as F
import dgl
import dgl.function as fn
from dgl.nn.pytorch import GATConv
from dgl.nn.pytorch.softmax import edge_softmax
from dgl.utils import expand_as_pair

from .base import GNNLayerBase, GNNBase
from ...data.data import GraphData
from ..utils.generic_utils import Identity


class GAT(GNNBase):
    # TODO: improve math descriptions bidirectional GNN.
    r"""Multi-layer Graph Attention Network (GAT).
    Support both `unidirectional GAT
    <https://arxiv.org/abs/1710.10903>`__ and bidirectional versions
    including `GAT-BiSep <https://arxiv.org/abs/1808.07624>`__ and
    `GAT-BiFuse <https://arxiv.org/abs/1908.04942>`__.

    .. math::
        h_i^{(l+1)} = \sum_{j\in \mathcal{N}(i)} \alpha_{i,j} W^{(l)} h_j^{(l)}
    where :math:`\alpha_{ij}` is the attention score bewteen node :math:`i` and
    node :math:`j`:

    .. math::
        \alpha_{ij}^{l} & = \mathrm{softmax_i} (e_{ij}^{l})
        e_{ij}^{l} & = \mathrm{LeakyReLU}\left(\vec{a}^T [W h_{i} \| W h_{j}]\right)

    Parameters
    ----------
    num_layers: int
        Number of GAT layers.
    input_size : int, or pair of ints
        Input feature size.
        If the layer is to be applied to a unidirectional bipartite graph, ``input_size``
        specifies the input feature size on both the source and destination nodes.  If
        a scalar is given, the source and destination node feature size would take the
        same value.
    hidden_size: int or list of int
        Hidden size per GAT layer. If ``int`` is given, all layers are forced to have the same hidden size.
    output_size : int
        Output feature size.
    heads : int or list of int
        Number of heads per GAT layer. If ``int`` is given, all layers are forced to have the same number of heads.
    direction_option: str
        Whether to use unidirectional (i.e., regular) or bidirectional (i.e., "bi_sep" and "bi_fuse") versions.
        Default : ``'bi_sep'``.
    feat_drop : float, optional
        Dropout rate on feature, default: ``0``.
    attn_drop : float, optional
        Dropout rate on attention weight, default: ``0``.
    negative_slope : float, optional
        LeakyReLU angle of negative slope, default: ``0.2``.
    residual : bool, optional
        If True, use residual connection.
        Default: ``False``.
    activation : callable activation function/layer or None, optional.
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    """
    def __init__(self,
                num_layers,
                input_size,
                hidden_size,
                output_size,
                heads,
                direction_option='bi_sep',
                feat_drop=0.,
                attn_drop=0.,
                negative_slope=0.2,
                residual=False,
                activation=None):
        super(GAT, self).__init__()
        self.num_layers = num_layers
        self.direction_option = direction_option
        self.gat_layers = nn.ModuleList()
        assert self.num_layers > 0
        if isinstance(hidden_size, int):
            hidden_size = [hidden_size] * (self.num_layers - 1)

        if isinstance(heads, int):
            heads = [heads] * self.num_layers

        if self.num_layers > 1:
            # input projection
            self.gat_layers.append(GATLayer(input_size,
                                            hidden_size[0],
                                            heads[0],
                                            direction_option=self.direction_option,
                                            feat_drop=feat_drop,
                                            attn_drop=attn_drop,
                                            negative_slope=negative_slope,
                                            residual=residual,
                                            activation=activation))

        # hidden layers
        for l in range(1, self.num_layers - 1):
            # due to multi-head, the input_size = hidden_size * num_heads
            self.gat_layers.append(GATLayer(hidden_size[l - 1] * heads[l - 1],
                                            hidden_size[l],
                                            heads[l],
                                            direction_option=self.direction_option,
                                            feat_drop=feat_drop,
                                            attn_drop=attn_drop,
                                            negative_slope=negative_slope,
                                            residual=residual,
                                            activation=activation))
        # output projection
        self.gat_layers.append(GATLayer(hidden_size[-1] * heads[-2] if self.num_layers > 1 else input_size,
                                        output_size,
                                        heads[-1],
                                        direction_option=self.direction_option,
                                        feat_drop=feat_drop,
                                        attn_drop=attn_drop,
                                        negative_slope=negative_slope,
                                        residual=residual,
                                        activation=None))

    def forward(self, graph):
        r"""Compute multi-layer graph attention network.

        Parameters
        ----------
        graph : GraphData
            The graph data containing topology and features.

        Returns
        -------
        GraphData
            The output graph data containing updated embeddings.
        """
        feat = graph.node_features['node_feat']
        dgl_graph = graph.to_dgl()

        if self.direction_option == 'bi_sep':
            h = [feat, feat]
        else:
            h = feat

        for l in range(self.num_layers - 1):
            h = self.gat_layers[l](dgl_graph, h)
            if self.direction_option == 'bi_sep':
                h = [each.flatten(1) for each in h]
            else:
                h = h.flatten(1)

        # output projection
        logits = self.gat_layers[-1](dgl_graph, h)

        if self.direction_option == 'bi_sep':
            logits = [each.mean(1) for each in logits]
            logits = torch.cat(logits, -1)
        else:
            logits = logits.mean(1)

        graph.node_features['node_emb'] = logits

        return graph

class GATLayer(GNNLayerBase):
    # TODO: improve math descriptions bidirectional GNN.
    r"""Single-layer Graph Attention Network (GAT).
    Support both `unidirectional GAT
    <https://arxiv.org/abs/1710.10903>`__ and bidirectional versions
    including `GAT-BiSep <https://arxiv.org/abs/1808.07624>`__ and
    `GAT-BiFuse <https://arxiv.org/abs/1908.04942>`__.

    .. math::
        h_i^{(l+1)} = \sum_{j\in \mathcal{N}(i)} \alpha_{i,j} W^{(l)} h_j^{(l)}
    where :math:`\alpha_{ij}` is the attention score bewteen node :math:`i` and
    node :math:`j`:

    .. math::
        \alpha_{ij}^{l} & = \mathrm{softmax_i} (e_{ij}^{l})
        e_{ij}^{l} & = \mathrm{LeakyReLU}\left(\vec{a}^T [W h_{i} \| W h_{j}]\right)

    Parameters
    ----------
    input_size : int, or pair of ints
        Input feature size.
        If the layer is to be applied to a unidirectional bipartite graph, ``input_size``
        specifies the input feature size on both the source and destination nodes.  If
        a scalar is given, the source and destination node feature size would take the
        same value.
    output_size : int
        Output feature size.
    num_heads : int
        Number of heads in Multi-Head Attention.
    direction_option: str
        Whether use unidirectional (i.e., regular) or bidirectional (i.e., `bi_sep` and `bi_fuse`) versions.
        Default: ``'bi_sep'``.
    feat_drop : float, optional
        Dropout rate on feature, default: ``0``.
    attn_drop : float, optional
        Dropout rate on attention weight, default: ``0``.
    negative_slope : float, optional
        LeakyReLU angle of negative slope, default: ``0.2``.
    residual : bool, optional
        If True, use residual connection.
        Default: ``False``.
    activation : callable activation function/layer or None, optional.
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    """
    def __init__(self,
                input_size,
                output_size,
                num_heads,
                direction_option='bi_sep',
                feat_drop=0.,
                attn_drop=0.,
                negative_slope=0.2,
                residual=False,
                activation=None):
        super(GATLayer, self).__init__()
        if direction_option == 'undirected':
            self.model = UndirectedGATLayerConv(input_size,
                                        output_size,
                                        num_heads,
                                        feat_drop=feat_drop,
                                        attn_drop=attn_drop,
                                        negative_slope=negative_slope,
                                        residual=residual,
                                        activation=activation)
        elif direction_option == 'bi_sep':
            self.model = BiSepGATLayerConv(input_size,
                                        output_size,
                                        num_heads,
                                        feat_drop=feat_drop,
                                        attn_drop=attn_drop,
                                        negative_slope=negative_slope,
                                        residual=residual,
                                        activation=activation)
        elif direction_option == 'bi_fuse':
            self.model = BiFuseGATLayerConv(input_size,
                                            output_size,
                                            num_heads,
                                            feat_drop=feat_drop,
                                            attn_drop=attn_drop,
                                            negative_slope=negative_slope,
                                            residual=residual,
                                            activation=activation)
        else:
            raise RuntimeError('Unknown `direction_option` value: {}'.format(direction_option))

    def forward(self, graph, feat):
        r"""Compute graph attention network layer.

        Parameters
        ----------
        graph : DGLGraph
            The graph.
        feat : torch.Tensor or pair of torch.Tensor
            If a torch.Tensor is given, the input feature of shape :math:`(N, D_{in})` where
            :math:`D_{in}` is size of input feature, :math:`N` is the number of nodes.
            If a pair of torch.Tensor is given, the pair must contain two tensors of shape
            :math:`(N_{in}, D_{in_{src}})` and :math:`(N_{out}, D_{in_{dst}})`.

        Returns
        -------
        torch.Tensor
            The output feature of shape :math:`(N, H, D_{out})` where :math:`H`
            is the number of heads, and :math:`D_{out}` is size of output feature.
        """
        return self.model(graph, feat)

class UndirectedGATLayerConv(GNNLayerBase):
    r"""Apply `Graph Attention Network <https://arxiv.org/abs/1710.10903>`__
    over an input signal.

    .. math::
        h_i^{(l+1)} = \sum_{j\in \mathcal{N}(i)} \alpha_{i,j} W^{(l)} h_j^{(l)}
    where :math:`\alpha_{ij}` is the attention score bewteen node :math:`i` and
    node :math:`j`:

    .. math::
        \alpha_{ij}^{l} & = \mathrm{softmax_i} (e_{ij}^{l})
        e_{ij}^{l} & = \mathrm{LeakyReLU}\left(\vec{a}^T [W h_{i} \| W h_{j}]\right)

    Parameters
    ----------
    input_size : int, or pair of ints
        Input feature size.
        If the layer is to be applied to a unidirectional bipartite graph, ``input_size``
        specifies the input feature size on both the source and destination nodes.  If
        a scalar is given, the source and destination node feature size would take the
        same value.
    output_size : int
        Output feature size.
    num_heads : int
        Number of heads in Multi-Head Attention.
    feat_drop : float, optional
        Dropout rate on feature, default: ``0``.
    attn_drop : float, optional
        Dropout rate on attention weight, default: ``0``.
    negative_slope : float, optional
        LeakyReLU angle of negative slope, default: ``0.2``.
    residual : bool, optional
        If True, use residual connection.
        Default: ``False``.
    activation : callable activation function/layer or None, optional.
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    """
    def __init__(self,
                input_size,
                output_size,
                num_heads,
                feat_drop=0.,
                attn_drop=0.,
                negative_slope=0.2,
                residual=False,
                activation=None):
        super(UndirectedGATLayerConv, self).__init__()
        self.model = GATConv(input_size, output_size, num_heads, feat_drop,
                            attn_drop, negative_slope, residual, activation)

    def forward(self, graph, feat):
        r"""Compute graph attention network layer.

        Parameters
        ----------
        graph : DGLGraph
            The graph.
        feat : torch.Tensor or pair of torch.Tensor
            If a torch.Tensor is given, the input feature of shape :math:`(N, D_{in})` where
            :math:`D_{in}` is size of input feature, :math:`N` is the number of nodes.
            If a pair of torch.Tensor is given, the pair must contain two tensors of shape
            :math:`(N_{in}, D_{in_{src}})` and :math:`(N_{out}, D_{in_{dst}})`.

        Returns
        -------
        torch.Tensor
            The output feature of shape :math:`(N, H, D_{out})` where :math:`H`
            is the number of heads, and :math:`D_{out}` is size of output feature.
        """
        return self.model(graph, feat)

class BiFuseGATLayerConv(GNNLayerBase):
    # TODO: improve math descriptions bidirectional GNN.
    r"""Apply `Bidirectional Fuse GNN mechanism <https://arxiv.org/abs/1908.04942>`__
    to `Graph Attention Network <https://arxiv.org/abs/1710.10903>`__ over an
    input signal. Fuse aggregated embeddings from both incoming and outgoing
    directions before updating node embeddings.

    .. math::
        h_i^{(l+1)} = \sum_{j\in \mathcal{N}(i)} \alpha_{i,j} W^{(l)} h_j^{(l)}
    where :math:`\alpha_{ij}` is the attention score bewteen node :math:`i` and
    node :math:`j`:

    .. math::
        \alpha_{ij}^{l} & = \mathrm{softmax_i} (e_{ij}^{l})
        e_{ij}^{l} & = \mathrm{LeakyReLU}\left(\vec{a}^T [W h_{i} \| W h_{j}]\right)

    Parameters
    ----------
    input_size : int, or pair of ints
        Input feature size.
        If the layer is to be applied to a unidirectional bipartite graph, ``input_size``
        specifies the input feature size on both the source and destination nodes.  If
        a scalar is given, the source and destination node feature size would take the
        same value.
    output_size : int
        Output feature size.
    num_heads : int
        Number of heads in Multi-Head Attention.
    feat_drop : float, optional
        Dropout rate on feature, default: ``0``.
    attn_drop : float, optional
        Dropout rate on attention weight, default: ``0``.
    negative_slope : float, optional
        LeakyReLU angle of negative slope, default: ``0.2``.
    residual : bool, optional
        If True, use residual connection.
        Default: ``False``.
    activation : callable activation function/layer or None, optional.
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    """
    def __init__(self,
                input_size,
                output_size,
                num_heads,
                feat_drop=0.,
                attn_drop=0.,
                negative_slope=0.2,
                residual=False,
                activation=None):
        super(BiFuseGATLayerConv, self).__init__()
        self._num_heads = num_heads
        self._in_src_feats, self._in_dst_feats = expand_as_pair(input_size)
        self._out_feats = output_size
        if isinstance(input_size, tuple):
            self.fc_src_fw = nn.Linear(
                self._in_src_feats, output_size * num_heads, bias=False)
            self.fc_dst_fw = nn.Linear(
                self._in_dst_feats, output_size * num_heads, bias=False)

            self.fc_src_bw = nn.Linear(
                self._in_src_feats, output_size * num_heads, bias=False)
            self.fc_dst_bw = nn.Linear(
                self._in_dst_feats, output_size * num_heads, bias=False)
        else:
            self.fc_fw = nn.Linear(
                self._in_src_feats, output_size * num_heads, bias=False)
            self.fc_bw = nn.Linear(
                self._in_src_feats, output_size * num_heads, bias=False)

        self.attn_l_fw = nn.Parameter(torch.FloatTensor(size=(1, num_heads, output_size)))
        self.attn_l_bw = nn.Parameter(torch.FloatTensor(size=(1, num_heads, output_size)))
        self.attn_r_fw = nn.Parameter(torch.FloatTensor(size=(1, num_heads, output_size)))
        self.attn_r_bw = nn.Parameter(torch.FloatTensor(size=(1, num_heads, output_size)))
        self.feat_drop = nn.Dropout(feat_drop)
        self.attn_drop = nn.Dropout(attn_drop)
        self.leaky_relu_fw = nn.LeakyReLU(negative_slope)
        self.leaky_relu_bw = nn.LeakyReLU(negative_slope)
        if residual:
            if self._in_dst_feats != output_size:
                self.res_fc = nn.Linear(
                    self._in_dst_feats, num_heads * output_size, bias=False)
            else:
                self.res_fc = Identity()
        else:
            self.register_buffer('res_fc', None)
        self.reset_parameters()
        self.activation = activation

        self.fuse_linear = nn.Linear(4 * output_size, output_size, bias=True)

    def reset_parameters(self):
        """Reinitialize learnable parameters."""
        gain = nn.init.calculate_gain('relu')
        if hasattr(self, 'fc_fw') or hasattr(self, 'fc_bw'):
            nn.init.xavier_normal_(self.fc_fw.weight, gain=gain)
            nn.init.xavier_normal_(self.fc_bw.weight, gain=gain)
        else: # bipartite graph neural works
            nn.init.xavier_normal_(self.fc_src_fw.weight, gain=gain)
            nn.init.xavier_normal_(self.fc_src_bw.weight, gain=gain)
            nn.init.xavier_normal_(self.fc_dst_fw.weight, gain=gain)
            nn.init.xavier_normal_(self.fc_dst_bw.weight, gain=gain)

        nn.init.xavier_normal_(self.attn_l_fw, gain=gain)
        nn.init.xavier_normal_(self.attn_l_bw, gain=gain)
        nn.init.xavier_normal_(self.attn_r_fw, gain=gain)
        nn.init.xavier_normal_(self.attn_r_bw, gain=gain)
        if isinstance(self.res_fc, nn.Linear):
            nn.init.xavier_normal_(self.res_fc.weight, gain=gain)

        if hasattr(self, 'fuse_linear'):
            nn.init.xavier_normal_(self.fuse_linear.weight, gain=gain)

    def forward(self, graph, feat):
        """Parameters
        ----------
        graph : DGLGraph
            The graph.
        feat : torch.Tensor or pair of torch.Tensor
            If a torch.Tensor is given, the input feature of shape :math:`(N, D_{in})` where
            :math:`D_{in}` is size of input feature, :math:`N` is the number of nodes.
            If a pair of torch.Tensor is given, the pair must contain two tensors of shape
            :math:`(N_{in}, D_{in_{src}})` and :math:`(N_{out}, D_{in_{dst}})`.

        Returns
        -------
        torch.Tensor
            The output feature of shape :math:`(N, H, D_{out})` where :math:`H`
            is the number of heads, and :math:`D_{out}` is size of output feature.
        """
        feat_fw = feat_bw = feat

        if isinstance(feat_fw, tuple):
            h_src_fw = self.feat_drop(feat_fw[0])
            h_dst_fw = self.feat_drop(feat_fw[1])
        else:
            h_src_fw = h_dst_fw = self.feat_drop(feat_fw)

        if isinstance(feat_bw, tuple):
            h_src_bw = self.feat_drop(feat_bw[0])
            h_dst_bw = self.feat_drop(feat_bw[1])
        else:
            h_src_bw = h_dst_bw = self.feat_drop(feat_bw)

        # forward direction
        with graph.local_scope():
            if isinstance(feat_fw, tuple):
                feat_src = self.fc_src_fw(h_src_fw).view(-1, self._num_heads, self._out_feats)
                feat_dst = self.fc_dst_fw(h_dst_fw).view(-1, self._num_heads, self._out_feats)
            else:
                feat_src = feat_dst = self.fc_fw(h_src_fw).view(
                    -1, self._num_heads, self._out_feats)
            # NOTE: GAT paper uses "first concatenation then linear projection"
            # to compute attention scores, while ours is "first projection then
            # addition", the two approaches are mathematically equivalent:
            # We decompose the weight vector a mentioned in the paper into
            # [a_l || a_r], then
            # a^T [Wh_i || Wh_j] = a_l Wh_i + a_r Wh_j
            # Our implementation is much efficient because we do not need to
            # save [Wh_i || Wh_j] on edges, which is not memory-efficient. Plus,
            # addition could be optimized with DGL's built-in function u_add_v,
            # which further speeds up computation and saves memory footprint.
            el = (feat_src * self.attn_l_fw).sum(dim=-1).unsqueeze(-1)
            er = (feat_dst * self.attn_r_fw).sum(dim=-1).unsqueeze(-1)
            graph.srcdata.update({'ft': feat_src, 'el': el})
            graph.dstdata.update({'er': er})
            # compute edge attention, el and er are a_l Wh_i and a_r Wh_j respectively.
            graph.apply_edges(fn.u_add_v('el', 'er', 'e'))
            e = self.leaky_relu_fw(graph.edata.pop('e'))
            # compute softmax
            graph.edata['a'] = self.attn_drop(edge_softmax(graph, e))
            # message passing
            graph.update_all(fn.u_mul_e('ft', 'a', 'm'),
                             fn.sum('m', 'ft'))
            agg_emb_fw = graph.dstdata['ft']


        # backward direction
        graph = graph.reverse()
        with graph.local_scope():
            if isinstance(feat_bw, tuple):
                feat_src = self.fc_src_bw(h_src_bw).view(-1, self._num_heads, self._out_feats)
                feat_dst = self.fc_dst_bw(h_dst_bw).view(-1, self._num_heads, self._out_feats)
            else:
                feat_src = feat_dst = self.fc_bw(h_src_bw).view(
                    -1, self._num_heads, self._out_feats)

            el = (feat_src * self.attn_l_bw).sum(dim=-1).unsqueeze(-1)
            er = (feat_dst * self.attn_r_bw).sum(dim=-1).unsqueeze(-1)
            graph.srcdata.update({'ft': feat_src, 'el': el})
            graph.dstdata.update({'er': er})
            # compute edge attention, el and er are a_l Wh_i and a_r Wh_j respectively.
            graph.apply_edges(fn.u_add_v('el', 'er', 'e'))
            e = self.leaky_relu_bw(graph.edata.pop('e'))
            # compute softmax
            graph.edata['a'] = self.attn_drop(edge_softmax(graph, e))
            # message passing
            graph.update_all(fn.u_mul_e('ft', 'a', 'm'),
                             fn.sum('m', 'ft'))
            agg_emb_bw = graph.dstdata['ft']

        fuse_vector = torch.cat(
            [agg_emb_fw, agg_emb_bw, agg_emb_fw * agg_emb_bw, agg_emb_fw - agg_emb_bw], dim=-1)
        fuse_gate_vector = torch.sigmoid(self.fuse_linear(fuse_vector))
        agg_emb = fuse_gate_vector * agg_emb_fw + (1 - fuse_gate_vector) * agg_emb_bw


        # residual
        if self.res_fc is not None:
            h_dst = h_dst_fw
            resval = self.res_fc(h_dst).view(h_dst.shape[0], -1, self._out_feats)
            agg_emb = agg_emb + resval

        # activation
        if self.activation:
            agg_emb = self.activation(agg_emb)

        return agg_emb

class BiSepGATLayerConv(GNNLayerBase):
    # TODO: improve math descriptions bidirectional GNN.
    r"""Apply `Bidirectional Separate GNN mechanism <https://arxiv.org/abs/1808.07624>`__
    to `Graph Attention Network <https://arxiv.org/abs/1710.10903>`__ over an
    input signal. Compute node embeddings for incoming and outgoing directions
    separately, and then concatenate the two output node embeddings
    after the final layer.

    .. math::
        h_i^{(l+1)} = \sum_{j\in \mathcal{N}(i)} \alpha_{i,j} W^{(l)} h_j^{(l)}
    where :math:`\alpha_{ij}` is the attention score bewteen node :math:`i` and
    node :math:`j`:

    .. math::
        \alpha_{ij}^{l} & = \mathrm{softmax_i} (e_{ij}^{l})
        e_{ij}^{l} & = \mathrm{LeakyReLU}\left(\vec{a}^T [W h_{i} \| W h_{j}]\right)

    Parameters
    ----------
    input_size : int, or pair of ints
        Input feature size.
        If the layer is to be applied to a unidirectional bipartite graph, ``input_size``
        specifies the input feature size on both the source and destination nodes.  If
        a scalar is given, the source and destination node feature size would take the
        same value.
    output_size : int
        Output feature size.
    num_heads : int
        Number of heads in Multi-Head Attention.
    feat_drop : float, optional
        Dropout rate on feature, default: ``0``.
    attn_drop : float, optional
        Dropout rate on attention weight, default: ``0``.
    negative_slope : float, optional
        LeakyReLU angle of negative slope, default: ``0.2``.
    residual : bool, optional
        If True, use residual connection.
        Default: ``False``.
    activation : callable activation function/layer or None, optional.
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    """
    def __init__(self,
                input_size,
                output_size,
                num_heads,
                feat_drop=0.,
                attn_drop=0.,
                negative_slope=0.2,
                residual=False,
                activation=None):
        super(BiSepGATLayerConv, self).__init__()
        self._num_heads = num_heads
        self._in_src_feats, self._in_dst_feats = expand_as_pair(input_size)
        self._out_feats = output_size
        if isinstance(input_size, tuple):
            self.fc_src_fw = nn.Linear(
                self._in_src_feats, output_size * num_heads, bias=False)
            self.fc_dst_fw = nn.Linear(
                self._in_dst_feats, output_size * num_heads, bias=False)

            self.fc_src_bw = nn.Linear(
                self._in_src_feats, output_size * num_heads, bias=False)
            self.fc_dst_bw = nn.Linear(
                self._in_dst_feats, output_size * num_heads, bias=False)
        else:
            self.fc_fw = nn.Linear(
                self._in_src_feats, output_size * num_heads, bias=False)
            self.fc_bw = nn.Linear(
                self._in_src_feats, output_size * num_heads, bias=False)

        self.attn_l_fw = nn.Parameter(torch.FloatTensor(size=(1, num_heads, output_size)))
        self.attn_l_bw = nn.Parameter(torch.FloatTensor(size=(1, num_heads, output_size)))
        self.attn_r_fw = nn.Parameter(torch.FloatTensor(size=(1, num_heads, output_size)))
        self.attn_r_bw = nn.Parameter(torch.FloatTensor(size=(1, num_heads, output_size)))
        self.feat_drop = nn.Dropout(feat_drop)
        self.attn_drop = nn.Dropout(attn_drop)
        self.leaky_relu_fw = nn.LeakyReLU(negative_slope)
        self.leaky_relu_bw = nn.LeakyReLU(negative_slope)
        if residual:
            if self._in_dst_feats != output_size:
                self.res_fc_fw = nn.Linear(
                    self._in_dst_feats, num_heads * output_size, bias=False)
                self.res_fc_bw = nn.Linear(
                    self._in_dst_feats, num_heads * output_size, bias=False)
            else:
                self.res_fc_fw = self.res_fc_bw = Identity()
        else:
            self.register_buffer('res_fc_fw', None)
            self.register_buffer('res_fc_bw', None)
        self.reset_parameters()
        self.activation = activation

    def reset_parameters(self):
        """Reinitialize learnable parameters."""
        gain = nn.init.calculate_gain('relu')
        if hasattr(self, 'fc_fw') or hasattr(self, 'fc_bw'):
            nn.init.xavier_normal_(self.fc_fw.weight, gain=gain)
            nn.init.xavier_normal_(self.fc_bw.weight, gain=gain)
        else: # bipartite graph neural works
            nn.init.xavier_normal_(self.fc_src_fw.weight, gain=gain)
            nn.init.xavier_normal_(self.fc_src_bw.weight, gain=gain)
            nn.init.xavier_normal_(self.fc_dst_fw.weight, gain=gain)
            nn.init.xavier_normal_(self.fc_dst_bw.weight, gain=gain)

        nn.init.xavier_normal_(self.attn_l_fw, gain=gain)
        nn.init.xavier_normal_(self.attn_l_bw, gain=gain)
        nn.init.xavier_normal_(self.attn_r_fw, gain=gain)
        nn.init.xavier_normal_(self.attn_r_bw, gain=gain)

        if isinstance(self.res_fc_fw, nn.Linear):
            nn.init.xavier_normal_(self.res_fc_fw.weight, gain=gain)

        if isinstance(self.res_fc_bw, nn.Linear):
            nn.init.xavier_normal_(self.res_fc_bw.weight, gain=gain)

        if hasattr(self, 'fuse_linear'):
            nn.init.xavier_normal_(self.fuse_linear.weight, gain=gain)

    def forward(self, graph, feat):
        r"""Compute graph attention network layer.

        Parameters
        ----------
        graph : DGLGraph
            The graph.
        feat:
            A list/tuple containing incoming feature ``feat_fw``
            and outgoing feature ``feat_bw``.
            - ``feat_fw`` : torch.Tensor or pair of torch.Tensor
                    If a torch.Tensor is given, the input feature of
                    shape :math:`(N, D_{in})` where :math:`D_{in}` is
                    size of input feature, :math:`N` is the number of
                    nodes. If a pair of torch.Tensor is given, the pair
                    must contain two tensors of shape :math:`(N_{in},
                    D_{in_{src}})` and :math:`(N_{out}, D_{in_{dst}})`.
            - ``feat_bw``: torch.Tensor or pair of torch.Tensor
                    If a torch.Tensor is given, the input feature of
                    shape :math:`(N, D_{in})` where :math:`D_{in}` is
                    size of input feature, :math:`N` is the number of
                    nodes. If a pair of torch.Tensor is given, the pair
                    must contain two tensors of shape :math:`(N_{in},
                    D_{in_{src}})` and :math:`(N_{out}, D_{in_{dst}})`.

        Returns
        -------
        Pair of torch.Tensor
            Each output feature of shape :math:`(N, H, D_{out})` where :math:`H`
            is the number of heads, and :math:`D_{out}` is size of output feature.
        """
        feat_fw, feat_bw = feat

        if isinstance(feat_fw, tuple):
            h_src_fw = self.feat_drop(feat_fw[0])
            h_dst_fw = self.feat_drop(feat_fw[1])
        else:
            h_src_fw = h_dst_fw = self.feat_drop(feat_fw)

        if isinstance(feat_bw, tuple):
            h_src_bw = self.feat_drop(feat_bw[0])
            h_dst_bw = self.feat_drop(feat_bw[1])
        else:
            h_src_bw = h_dst_bw = self.feat_drop(feat_bw)


        # forward direction
        with graph.local_scope():
            if isinstance(feat_fw, tuple):
                feat_src = self.fc_src_fw(h_src_fw).view(-1, self._num_heads, self._out_feats)
                feat_dst = self.fc_dst_fw(h_dst_fw).view(-1, self._num_heads, self._out_feats)
            else:
                feat_src = feat_dst = self.fc_fw(h_src_fw).view(
                    -1, self._num_heads, self._out_feats)
            # NOTE: GAT paper uses "first concatenation then linear projection"
            # to compute attention scores, while ours is "first projection then
            # addition", the two approaches are mathematically equivalent:
            # We decompose the weight vector a mentioned in the paper into
            # [a_l || a_r], then
            # a^T [Wh_i || Wh_j] = a_l Wh_i + a_r Wh_j
            # Our implementation is much efficient because we do not need to
            # save [Wh_i || Wh_j] on edges, which is not memory-efficient. Plus,
            # addition could be optimized with DGL's built-in function u_add_v,
            # which further speeds up computation and saves memory footprint.
            el = (feat_src * self.attn_l_fw).sum(dim=-1).unsqueeze(-1)
            er = (feat_dst * self.attn_r_fw).sum(dim=-1).unsqueeze(-1)
            graph.srcdata.update({'ft': feat_src, 'el': el})
            graph.dstdata.update({'er': er})
            # compute edge attention, el and er are a_l Wh_i and a_r Wh_j respectively.
            graph.apply_edges(fn.u_add_v('el', 'er', 'e'))
            e = self.leaky_relu_fw(graph.edata.pop('e'))
            # compute softmax
            graph.edata['a'] = self.attn_drop(edge_softmax(graph, e))
            # message passing
            graph.update_all(fn.u_mul_e('ft', 'a', 'm'),
                             fn.sum('m', 'ft'))
            agg_emb_fw = graph.dstdata['ft']


        # backward direction
        graph = graph.reverse()
        with graph.local_scope():
            if isinstance(feat_bw, tuple):
                feat_src = self.fc_src_bw(h_src_bw).view(-1, self._num_heads, self._out_feats)
                feat_dst = self.fc_dst_bw(h_dst_bw).view(-1, self._num_heads, self._out_feats)
            else:
                feat_src = feat_dst = self.fc_bw(h_src_bw).view(
                    -1, self._num_heads, self._out_feats)

            el = (feat_src * self.attn_l_bw).sum(dim=-1).unsqueeze(-1)
            er = (feat_dst * self.attn_r_bw).sum(dim=-1).unsqueeze(-1)
            graph.srcdata.update({'ft': feat_src, 'el': el})
            graph.dstdata.update({'er': er})
            # compute edge attention, el and er are a_l Wh_i and a_r Wh_j respectively.
            graph.apply_edges(fn.u_add_v('el', 'er', 'e'))
            e = self.leaky_relu_bw(graph.edata.pop('e'))
            # compute softmax
            graph.edata['a'] = self.attn_drop(edge_softmax(graph, e))
            # message passing
            graph.update_all(fn.u_mul_e('ft', 'a', 'm'),
                             fn.sum('m', 'ft'))
            agg_emb_bw = graph.dstdata['ft']


        # residual
        if self.res_fc_fw is not None:
            resval_fw = self.res_fc_fw(h_dst_fw).view(h_dst_fw.shape[0], -1, self._out_feats)
            resval_bw = self.res_fc_bw(h_dst_bw).view(h_dst_bw.shape[0], -1, self._out_feats)
            agg_emb_fw = agg_emb_fw + resval_fw
            agg_emb_bw = agg_emb_bw + resval_bw

        # activation
        if self.activation:
            agg_emb_fw = self.activation(agg_emb_fw)
            agg_emb_bw = self.activation(agg_emb_bw)

        return [agg_emb_fw, agg_emb_bw]



from nltk.tokenize import word_tokenize
import torch
from torch import nn

from ...data.data import GraphData
from .base import DynamicGraphConstructionBase
from ..utils.generic_utils import normalize_adj, to_cuda
# from ..utils.constants import VERY_SMALL_NUMBER
from .utils import convert_adj_to_graph
from ...data.data import to_batch
from ..utils.vocab_utils import Vocab


class NodeEmbeddingBasedGraphConstruction(DynamicGraphConstructionBase):
    """Class for node embedding based dynamic graph construction.

    Parameters
    ----------
    word_vocab : Vocab
        The word vocabulary.
    embedding_styles : dict
        - ``single_token_item`` : specify whether the item (i.e., node or edge) contains single token or multiple tokens.
        - ``emb_strategy`` : specify the embedding construction strategy.
        - ``num_rnn_layers``: specify the number of RNN layers.
        - ``bert_model_name``: specify the BERT model name.
        - ``bert_lower_case``: specify whether to lower case the input text for BERT embeddings.
    """
    def __init__(self, word_vocab, embedding_styles, **kwargs):
        super(NodeEmbeddingBasedGraphConstruction, self).__init__(
                                                            word_vocab,
                                                            embedding_styles,
                                                            **kwargs)

    def forward(self, batch_graphdata):
        """Compute graph topology and initial node embeddings.

        Parameters
        ----------
        batch_graphdata : GraphData
            The input graph data.

        Returns
        -------
        GraphData
            The constructed batched graph.
        """
        batch_graphdata = self.embedding(batch_graphdata)
        batch_graphdata = self.topology(batch_graphdata)

        return batch_graphdata

    def topology(self, graph):
        """Compute graph topology.

        Parameters
        ----------
        graph : GraphData
            The input graph data.

        Returns
        -------
        GraphData
            The constructed graph.
        """
        node_emb = graph.batch_node_features["node_feat"]
        node_mask = (graph.batch_node_features["token_id"] != Vocab.PAD)

        raw_adj = self.compute_similarity_metric(node_emb, node_mask)
        raw_adj = self.sparsify_graph(raw_adj)
        graph_reg = self.compute_graph_regularization(raw_adj, node_emb)

        if self.sim_metric_type in ('rbf_kernel', 'weighted_cosine'):
            assert raw_adj.min().item() >= 0, 'adjacency matrix must be non-negative!'
            adj = raw_adj / torch.clamp(torch.sum(raw_adj, dim=-1, keepdim=True), min=torch.finfo(torch.float32).eps)
            reverse_adj = raw_adj / torch.clamp(torch.sum(raw_adj, dim=-2, keepdim=True), min=torch.finfo(torch.float32).eps)
        elif self.sim_metric_type == 'cosine':
            raw_adj = (raw_adj > 0).float()
            adj = normalize_adj(raw_adj)
            reverse_adj = adj
        else:
            adj = torch.softmax(raw_adj, dim=-1)
            reverse_adj = torch.softmax(raw_adj, dim=-2)

        graph = convert_adj_to_graph(graph, adj, reverse_adj, 0)
        graph.graph_attributes['graph_reg'] = graph_reg

        return graph

    @classmethod
    def init_topology(cls, raw_text_data, lower_case=True, tokenizer=word_tokenize):
        """Convert raw text data to the initial node set graph (i.e., no edge information).

        Parameters
        ----------
        raw_text_data : str or list/tuple of str
            The raw text data. When a list/tuple of tokens is provided, no
            tokenization will be conducted and each token is a node;
            otherwise, tokenization will be conducted on the input string
            to get a list of tokens.
        lower_case : boolean
            Specify whether to lower case the input text, default: ``True``.
        tokenizer : callable, optional
            The tokenization function.

        Returns
        -------
        GraphData
            The constructed graph.
        """
        if isinstance(raw_text_data, str):
            token_list = tokenizer(raw_text_data.strip())
        elif isinstance(raw_text_data, (list, tuple)):
            token_list = raw_text_data
        else:
            raise RuntimeError('raw_text_data must be str or list/tuple of str')

        graph = GraphData()
        graph.add_nodes(len(token_list))

        for idx in range(len(token_list)):
            graph.node_attributes[idx]['token'] = token_list[idx].lower() if lower_case else token_list[idx]

        return graph
